import inspect
from typing import Optional
from unittest import TestCase

from pcs.cli.reports import messages as cli_messages
from pcs.common import file_type_codes
from pcs.common.reports import (
    codes,
    const,
    item,
    messages,
    types,
)


class AllClassesTested(TestCase):
    def test_success(self):
        self.maxDiff = None
        message_classes = frozenset(
            name
            for name, member in inspect.getmembers(
                cli_messages, inspect.isclass
            )
            if issubclass(member, cli_messages.CliReportMessageCustom)
            and member != cli_messages.CliReportMessageCustom
        )
        test_classes = frozenset(
            name
            for name, member in inspect.getmembers(
                inspect.getmodule(self), inspect.isclass
            )
            if issubclass(member, CliReportMessageTestBase)
        )
        untested = sorted(message_classes - test_classes)
        self.assertEqual(
            untested,
            [],
            f"It seems {len(untested)} subclass(es) of 'CliReportMessageCustom' "
            "are missing tests. Make sure the test classes have the same name "
            "as the code classes.",
        )


class CliReportMessageTestBase(TestCase):
    @staticmethod
    def _get_cli_msg_obj(
        msg_obj: item.ReportItemMessage,
    ) -> cli_messages.CliReportMessage:
        return cli_messages.report_item_msg_from_dto(msg_obj.to_dto())

    def assert_message(
        self, msg_obj: item.ReportItemMessage, expected_msg: str
    ) -> None:
        self.assertEqual(self._get_cli_msg_obj(msg_obj).message, expected_msg)
        self.assert_same_messages_with_force_code(msg_obj, None)

    def assert_message_with_force_code(
        self,
        msg_obj: item.ReportItemMessage,
        force_code: types.ForceCode,
        expected_msg: str,
    ) -> None:
        self.assertEqual(
            self._get_cli_msg_obj(msg_obj).get_message_with_force_text(
                force_code
            ),
            expected_msg,
        )

    def assert_same_messages_with_force_code(
        self,
        msg_obj: item.ReportItemMessage,
        force_code: Optional[types.ForceCode],
    ) -> None:
        cli_msg_obj = self._get_cli_msg_obj(msg_obj)
        self.assertEqual(
            cli_msg_obj.get_message_with_force_text(force_code),
            cli_msg_obj.message,
        )


class ResourceManagedNoMonitorEnabled(CliReportMessageTestBase):
    def test_message(self):
        self.assert_message(
            messages.ResourceManagedNoMonitorEnabled("resId"),
            (
                "Resource 'resId' has no enabled monitor operations. Re-run "
                "with '--monitor' to enable them."
            ),
        )


class ResourceUnmoveUnbanPcmkExpiredNotSupported(CliReportMessageTestBase):
    def test_message(self):
        self.assert_message(
            messages.ResourceUnmoveUnbanPcmkExpiredNotSupported(),
            "--expired not supported, please upgrade pacemaker",
        )


class CannotUnmoveUnbanResourceMasterResourceNotPromotable(
    CliReportMessageTestBase
):
    def test_with_promotable_id(self):
        self.assert_message(
            messages.CannotUnmoveUnbanResourceMasterResourceNotPromotable(
                "R", "P"
            ),
            (
                "when specifying --master you must use the promotable clone id "
                "(P)"
            ),
        )

    def test_without_promotable_id(self):
        self.assert_message(
            messages.CannotUnmoveUnbanResourceMasterResourceNotPromotable("R"),
            "when specifying --master you must use the promotable clone id",
        )


class CannotBanResourceMasterResourceNotPromotable(CliReportMessageTestBase):
    def test_with_promotable_id(self):
        self.assert_message(
            messages.CannotBanResourceMasterResourceNotPromotable("R", "P"),
            (
                "when specifying --master you must use the promotable clone id "
                "(P)"
            ),
        )

    def test_without_promotable_id(self):
        self.assert_message(
            messages.CannotBanResourceMasterResourceNotPromotable("R"),
            "when specifying --master you must use the promotable clone id",
        )


class CannotMoveResourceMasterResourceNotPromotable(CliReportMessageTestBase):
    def test_with_promotable_id(self):
        self.assert_message(
            messages.CannotMoveResourceMasterResourceNotPromotable("R", "P"),
            (
                "when specifying --master you must use the promotable clone id "
                "(P)"
            ),
        )

    def test_without_promotable_id(self):
        self.assert_message(
            messages.CannotMoveResourceMasterResourceNotPromotable("R"),
            "when specifying --master you must use the promotable clone id",
        )


class CannotMoveResourceNotRunning(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.CannotMoveResourceNotRunning("R"),
            (
                "It is not possible to move resource 'R' as it is not running "
                "at the moment, use 'pcs resource move' without "
                "'--autodelete' option or 'pcs constraint location' commands "
                "instead"
            ),
        )


class InvalidCibContent(CliReportMessageTestBase):
    def test_message_can_be_more_verbose(self):
        report = "no verbose\noutput\n"
        self.assert_message(
            messages.InvalidCibContent(report, True),
            "invalid cib:\n{0}\n\nUse --full for more details.".format(report),
        )

    def test_message_cannot_be_more_verbose(self):
        report = "some verbose\noutput"
        self.assert_message(
            messages.InvalidCibContent(report, False),
            "invalid cib:\n{0}".format(report),
        )


class NodeCommunicationErrorNotAuthorized(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.NodeCommunicationErrorNotAuthorized(
                "node1", "some-command", "some error"
            ),
            (
                "Unable to authenticate to node1 (some error), try running "
                "'pcs host auth node1'"
            ),
        )


class NodeCommunicationErrorTimedOut(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.NodeCommunicationErrorTimedOut(
                "node-1",
                "/remote/command",
                "Connection timed out after 60049 milliseconds",
            ),
            (
                "node-1: Connection timeout, check if pcsd is running there or "
                "try setting higher timeout with --request-timeout option "
                "(Connection timed out after 60049 milliseconds)"
            ),
        )


class SbdWatchdogTestMultipleDevices(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.SbdWatchdogTestMultipleDevices(),
            (
                "Multiple watchdog devices available, therefore, watchdog "
                "which should be tested has to be specified. To list available "
                "watchdog devices use command 'pcs stonith sbd watchdog list'"
            ),
        )


class NodeUsedAsTieBreaker(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.NodeUsedAsTieBreaker("node-1", "0"),
            (
                "Node 'node-1' with id '0' is used as a tie breaker for a "
                "qdevice, run 'pcs quorum device update model "
                "tie_breaker=<node id>' to change it"
            ),
        )


class NodesToRemoveUnreachable(CliReportMessageTestBase):
    def test_one_node(self):
        self.assert_message(
            messages.NodesToRemoveUnreachable(["node1"]),
            (
                "Removed node 'node1' could not be reached and subsequently "
                "deconfigured. Run 'pcs cluster destroy' on the unreachable "
                "node."
            ),
        )

    def test_multiple_nodes(self):
        self.assert_message(
            messages.NodesToRemoveUnreachable(["node2", "node1"]),
            (
                "Removed nodes 'node1', 'node2' could not be reached and "
                "subsequently deconfigured. Run 'pcs cluster destroy' on the "
                "unreachable nodes."
            ),
        )


class UnableToConnectToAllRemainingNodes(CliReportMessageTestBase):
    def test_one_node(self):
        self.assert_message(
            messages.UnableToConnectToAllRemainingNodes(["node1"]),
            (
                "Remaining cluster node 'node1' could not be reached, run "
                "'pcs cluster sync' on any currently online node once the "
                "unreachable one become available"
            ),
        )

    def test_multiple_nodes(self):
        self.assert_message(
            messages.UnableToConnectToAllRemainingNodes(["node2", "node1"]),
            (
                "Remaining cluster nodes 'node1', 'node2' could not be "
                "reached, run 'pcs cluster sync' on any currently online node "
                "once the unreachable ones become available"
            ),
        )


class CannotRemoveAllClusterNodes(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.CannotRemoveAllClusterNodes(),
            (
                "No nodes would be left in the cluster, if you intend to "
                "destroy the whole cluster, run 'pcs cluster destroy --all' "
                "instead"
            ),
        )


class WaitForNodeStartupWithoutStart(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.WaitForNodeStartupWithoutStart(),
            "Cannot specify '--wait' without specifying '--start'",
        )


class HostNotFound(CliReportMessageTestBase):
    def test_one_node(self):
        self.assert_message(
            messages.HostNotFound(["node1"]),
            (
                "Host 'node1' is not known to pcs, try to "
                "authenticate the host using 'pcs host auth node1' "
                "command"
            ),
        )

    def test_multiple_nodes(self):
        self.assert_message(
            messages.HostNotFound(["node2", "node1"]),
            (
                "Hosts 'node1', 'node2' are not known to pcs, try to "
                "authenticate the hosts using 'pcs host auth node1 node2' "
                "command"
            ),
        )


class UseCommandNodeRemoveGuest(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.UseCommandNodeRemoveGuest(),
            (
                "this command is not sufficient for removing a guest node, use"
                " 'pcs cluster node remove-guest'"
            ),
        )


class UseCommandNodeAddGuest(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.UseCommandNodeAddGuest(),
            (
                "this command is not sufficient for creating a guest node, use"
                " 'pcs cluster node add-guest'"
            ),
        )


class UseCommandNodeAddRemote(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.UseCommandNodeAddRemote(),
            (
                "this command is not sufficient for creating a remote "
                "connection, use 'pcs cluster node add-remote'"
            ),
        )


class CorosyncNodeConflictCheckSkipped(CliReportMessageTestBase):
    def test_not_live_cib(self):
        self.assert_message(
            messages.CorosyncNodeConflictCheckSkipped(
                const.REASON_NOT_LIVE_CIB
            ),
            (
                "Unable to check if there is a conflict with nodes set in "
                "corosync because the command does not run on a live cluster "
                "(e.g. -f "
                "was used)"
            ),
        )

    def test_unreachable(self):
        self.assert_message(
            messages.CorosyncNodeConflictCheckSkipped(const.REASON_UNREACHABLE),
            (
                "Unable to check if there is a conflict with nodes set in "
                "corosync because pcs is unable to connect to the node(s)"
            ),
        )


class LiveEnvironmentNotConsistent(CliReportMessageTestBase):
    def test_one_one(self):
        self.assert_message(
            messages.LiveEnvironmentNotConsistent(
                [file_type_codes.BOOTH_CONFIG],
                [file_type_codes.BOOTH_KEY],
            ),
            (
                "When '--booth-conf' is specified, "
                "'--booth-key' must be specified as well"
            ),
        )

    def test_many_many(self):
        self.assert_message(
            messages.LiveEnvironmentNotConsistent(
                [file_type_codes.CIB, file_type_codes.BOOTH_CONFIG],
                [file_type_codes.COROSYNC_CONF, file_type_codes.BOOTH_KEY],
            ),
            (
                "When '--booth-conf', '-f' are specified, "
                "'--booth-key', '--corosync_conf' must be specified as well"
            ),
        )


class LiveEnvironmentRequired(CliReportMessageTestBase):
    def test_build_messages_transformable_codes(self):
        self.assert_message(
            messages.LiveEnvironmentRequired(
                [file_type_codes.COROSYNC_CONF, file_type_codes.CIB]
            ),
            "This command does not support '--corosync_conf', '-f'",
        )


class LiveEnvironmentRequiredForLocalNode(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.LiveEnvironmentRequiredForLocalNode(),
            "Node(s) must be specified if -f is used",
        )


class ServiceCommandsOnNodesSkipped(CliReportMessageTestBase):
    def test_not_live(self):
        self.assert_message(
            messages.ServiceCommandsOnNodesSkipped(
                const.REASON_NOT_LIVE_CIB,
                ["pacemaker_remote enable", "pacemaker_remote start"],
                ["nodeA", "nodeB"],
            ),
            "Running action(s) 'pacemaker_remote enable', 'pacemaker_remote "
            "start' on 'nodeA', 'nodeB' was skipped because the command "
            "does not run on a live cluster (e.g. -f was used). Please, "
            "run the action(s) manually.",
        )

    def test_unreachable(self):
        self.assert_message(
            messages.ServiceCommandsOnNodesSkipped(
                const.REASON_UNREACHABLE,
                ["pacemaker_remote enable", "pacemaker_remote start"],
                ["nodeA", "nodeB"],
            ),
            "Running action(s) 'pacemaker_remote enable', 'pacemaker_remote "
            "start' on 'nodeA', 'nodeB' was skipped because pcs is unable "
            "to connect to the node(s). Please, run the action(s) manually.",
        )

    def test_unknown_reason(self):
        self.assert_message(
            messages.ServiceCommandsOnNodesSkipped(
                "some undefined reason",
                ["pacemaker_remote enable", "pacemaker_remote start"],
                ["nodeA", "nodeB"],
            ),
            "Running action(s) 'pacemaker_remote enable', 'pacemaker_remote "
            "start' on 'nodeA', 'nodeB' was skipped because some undefined "
            "reason. Please, run the action(s) manually.",
        )


class FilesRemoveFromNodesSkipped(CliReportMessageTestBase):
    def test_not_live(self):
        self.assert_message(
            messages.FilesRemoveFromNodesSkipped(
                const.REASON_NOT_LIVE_CIB, ["file1"], ["nodeA", "nodeB"]
            ),
            "Removing 'file1' from 'nodeA', 'nodeB' was skipped because the "
            "command does not run on a live cluster (e.g. -f was used). "
            "Please, remove the file(s) manually.",
        )

    def test_unreachable(self):
        self.assert_message(
            messages.FilesRemoveFromNodesSkipped(
                const.REASON_UNREACHABLE, ["file1", "file2"], ["nodeA"]
            ),
            "Removing 'file1', 'file2' from 'nodeA' was skipped because pcs is "
            "unable to connect to the node(s). Please, remove the file(s) "
            "manually.",
        )

    def test_unknown_reason(self):
        self.assert_message(
            messages.FilesRemoveFromNodesSkipped(
                "some undefined reason", ["file1", "file2"], ["nodeA", "nodeB"]
            ),
            "Removing 'file1', 'file2' from 'nodeA', 'nodeB' was skipped "
            "because some undefined reason. Please, remove the file(s) "
            "manually.",
        )


class FilesDistributionSkipped(CliReportMessageTestBase):
    def test_not_live(self):
        self.assert_message(
            messages.FilesDistributionSkipped(
                const.REASON_NOT_LIVE_CIB, ["file1"], ["nodeA", "nodeB"]
            ),
            "Distribution of 'file1' to 'nodeA', 'nodeB' was skipped because "
            "the command does not run on a live cluster (e.g. -f was used). "
            "Please, distribute the file(s) manually.",
        )

    def test_unreachable(self):
        self.assert_message(
            messages.FilesDistributionSkipped(
                const.REASON_UNREACHABLE, ["file1", "file2"], ["nodeA"]
            ),
            "Distribution of 'file1', 'file2' to 'nodeA' was skipped because "
            "pcs is unable to connect to the node(s). Please, distribute "
            "the file(s) manually.",
        )

    def test_unknown_reason(self):
        self.assert_message(
            messages.FilesDistributionSkipped(
                "some undefined reason", ["file1", "file2"], ["nodeA", "nodeB"]
            ),
            "Distribution of 'file1', 'file2' to 'nodeA', 'nodeB' was skipped "
            "because some undefined reason. Please, distribute the file(s) "
            "manually.",
        )


class WaitForIdleNotLiveCluster(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.WaitForIdleNotLiveCluster(),
            "Cannot use '-f' together with '--wait'",
        )


class TagCannotRemoveReferencesWithoutRemovingTag(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.TagCannotRemoveReferencesWithoutRemovingTag("tag-id"),
            (
                "There would be no references left in the tag 'tag-id', please "
                "remove the whole tag using the 'pcs tag remove tag-id' command"
            ),
        )


class RuleExpressionParseError(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.RuleExpressionParseError(
                "resource dummy op monitor",
                "Expected end of text",
                "resource dummy op monitor",
                1,
                16,
                15,
            ),
            "'resource dummy op monitor' is not a valid rule expression, "
            "parse error near or after line 1 column 16\n"
            "  resource dummy op monitor\n"
            "  ---------------^",
        )


class CibNvsetAmbiguousProvideNvsetId(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.CibNvsetAmbiguousProvideNvsetId(
                const.PCS_COMMAND_RESOURCE_DEFAULTS_UPDATE
            ),
            "Several options sets exist, please use the 'pcs resource defaults "
            "set update' command and specify an option set ID",
        )


class UnableToGetAgentMetadata(CliReportMessageTestBase):
    def test_all(self):
        self.assert_message(
            messages.UnableToGetAgentMetadata(
                "agent-name", "reason\non multiple lines"
            ),
            "Agent 'agent-name' is not installed or does not provide valid "
            "metadata: reason, on multiple lines",
        )


class InvalidResourceAgentName(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.InvalidResourceAgentName(":name"),
            "Invalid resource agent name ':name'. Use standard:provider:type "
            "when standard is 'ocf' or standard:type otherwise. List of standards "
            "and providers can be obtained by using commands 'pcs resource "
            "standards' and 'pcs resource providers'.",
        )


class InvalidStonithAgentName(CliReportMessageTestBase):
    def test_success(self):
        self.assert_message(
            messages.InvalidStonithAgentName("fence:name"),
            "Invalid stonith agent name 'fence:name'. Agent name cannot contain "
            "the ':' character, do not use the 'stonith:' prefix. List of agents "
            "can be obtained by using command 'pcs stonith list'.",
        )


class HostAlreadyInClusterConfig(CliReportMessageTestBase):
    def test_message(self):
        self.assert_message(
            messages.HostAlreadyInClusterConfig("nodeX"),
            "nodeX: The host seems to be in a cluster already as cluster "
            "configuration files have been found on the host. If the host is "
            "not part of a cluster, run 'pcs cluster destroy' on host 'nodeX' "
            "to remove those configuration files",
        )


class CannotLeaveGroupEmptyAfterMove(CliReportMessageTestBase):
    def test_single_resource(self):
        self.assert_message(
            messages.CannotLeaveGroupEmptyAfterMove("gr1", ["R"]),
            "Unable to move resource 'R' as it would leave group 'gr1' empty. "
            "Please, use the 'pcs resource ungroup gr1' command first.",
        )

    def test_multiple_resources(self):
        self.assert_message(
            messages.CannotLeaveGroupEmptyAfterMove("gr1", ["R3", "R1", "R2"]),
            "Unable to move resources 'R1', 'R2', 'R3' as it would leave "
            "group 'gr1' empty. Please, use the 'pcs resource ungroup gr1' "
            "command first.",
        )


class StonithRestartlessUpdateUnableToPerform(CliReportMessageTestBase):
    def test_other_reason(self):
        report_msg = messages.StonithRestartlessUpdateUnableToPerform("reason")
        self.assert_message(
            report_msg,
            report_msg.message,
        )

    def test_not_running_reason(self):
        report_msg = messages.StonithRestartlessUpdateUnableToPerform(
            "reason",
            const.STONITH_RESTARTLESS_UPDATE_UNABLE_TO_PERFORM_REASON_NOT_RUNNING,
        )
        self.assert_message(
            report_msg,
            f"{report_msg.message}, please use command 'pcs stonith update' instead",
        )


class AgentSelfValidationResult(CliReportMessageTestBase):
    def test_message(self):
        self.assert_message(
            messages.AgentSelfValidationResult("line 1\nline 3\nline 2"),
            "Validation result from agent:\n  line 1\n  line 3\n  line 2",
        )

    def test_force(self):
        self.assert_message_with_force_code(
            messages.AgentSelfValidationResult("line 1\nline 3\nline 2"),
            codes.FORCE,
            "Validation result from agent (use --force to override):\n  line 1\n  line 3\n  line 2",
        )

    def test_unknown_force_code(self):
        self.assert_same_messages_with_force_code(
            messages.AgentSelfValidationResult("line 1\nline 3\nline 2"),
            codes.SKIP_OFFLINE_NODES,
        )



import json
from typing import (
    Any,
    Dict,
    Optional,
    Type,
)

from dataclasses import dataclass

from pcs.common import file_type_codes as code
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.booth.config_facade import ConfigFacade as BoothConfigFacade
from pcs.lib.booth.config_parser import Exporter as BoothConfigExporter
from pcs.lib.booth.config_parser import Parser as BoothConfigParser
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.corosync.config_parser import Exporter as CorosyncConfigExporter
from pcs.lib.corosync.config_parser import Parser as CorosyncConfigParser
from pcs.lib.dr.config.facade import Facade as DrConfigFacade
from pcs.lib.interface.config import (
    ExporterInterface,
    FacadeInterface,
    ParserErrorException,
    ParserInterface,
)


@dataclass(frozen=True)
class FileToolbox:
    # File type code the toolbox belongs to
    file_type_code: code.FileTypeCode
    # Provides an easy access for reading and modifying data
    facade: Type[FacadeInterface]
    # Turns raw data into a structure which the facade is able to process
    parser: Type[ParserInterface]
    # Turns a structure produced by the parser and the facade to raw data
    exporter: Type[ExporterInterface]
    # Checks that the structure is valid
    validator: None  # TBI
    # Provides means for file syncing based on the file's version
    version_controller: None  # TBI


class JsonParserException(ParserErrorException):
    def __init__(self, json_exception: json.JSONDecodeError):
        super().__init__()
        self.json_exception = json_exception


class JsonParser(ParserInterface):
    """
    Adapts standard json parser to our interfaces
    """

    @staticmethod
    def parse(raw_file_data: bytes) -> Dict[str, Any]:
        try:
            # json.loads handles bytes, it expects utf-8, 16 or 32 encoding
            return json.loads(raw_file_data)
        except json.JSONDecodeError as e:
            raise JsonParserException(e) from e

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: code.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        if isinstance(exception, JsonParserException):
            if isinstance(exception.json_exception, json.JSONDecodeError):
                return [
                    ReportItem(
                        severity=reports.item.get_severity(
                            force_code,
                            is_forced_or_warning,
                        ),
                        message=reports.messages.ParseErrorJsonFile(
                            file_type_code,
                            exception.json_exception.lineno,
                            exception.json_exception.colno,
                            exception.json_exception.pos,
                            exception.json_exception.msg,
                            str(exception.json_exception),
                            file_path=file_path,
                        ),
                    )
                ]
        raise exception


class JsonExporter(ExporterInterface):
    """
    Adapts standard json exporter to our interfaces
    """

    @staticmethod
    def export(config_structure: Dict[str, Any]) -> bytes:
        return json.dumps(
            config_structure,
            indent=4,
            sort_keys=True,
        ).encode("utf-8")


class NoopParser(ParserInterface):
    @staticmethod
    def parse(raw_file_data: bytes) -> bytes:
        return raw_file_data

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: code.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        return []


class NoopExporter(ExporterInterface):
    @staticmethod
    def export(config_structure: bytes) -> bytes:
        return config_structure


class NoopFacade(FacadeInterface):
    @classmethod
    def create(cls) -> "NoopFacade":
        return cls(bytes())


_toolboxes = {
    code.BOOTH_CONFIG: FileToolbox(
        file_type_code=code.BOOTH_CONFIG,
        facade=BoothConfigFacade,
        parser=BoothConfigParser,
        exporter=BoothConfigExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.BOOTH_KEY: FileToolbox(
        file_type_code=code.BOOTH_KEY,
        facade=NoopFacade,
        parser=NoopParser,
        exporter=NoopExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.COROSYNC_CONF: FileToolbox(
        file_type_code=code.COROSYNC_CONF,
        facade=CorosyncConfigFacade,
        parser=CorosyncConfigParser,
        exporter=CorosyncConfigExporter,
        validator=None,
        version_controller=None,
    ),
    code.PACEMAKER_AUTHKEY: FileToolbox(
        file_type_code=code.PACEMAKER_AUTHKEY,
        facade=NoopFacade,
        parser=NoopParser,
        exporter=NoopExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_KNOWN_HOSTS: FileToolbox(
        file_type_code=code.PCS_KNOWN_HOSTS,
        # TODO needed for 'auth' and 'deauth' commands
        facade=None,  # type: ignore
        parser=JsonParser,
        exporter=JsonExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
    code.PCS_DR_CONFIG: FileToolbox(
        file_type_code=code.PCS_DR_CONFIG,
        facade=DrConfigFacade,
        parser=JsonParser,
        exporter=JsonExporter,
        validator=None,  # TODO needed for files syncing
        version_controller=None,  # TODO needed for files syncing
    ),
}


def for_file_type(file_type_code: code.FileTypeCode) -> FileToolbox:
    return _toolboxes[file_type_code]



from unittest import TestCase

from pcs.lib.corosync import config_validators

from pcs_test.tier0.lib.corosync.test_config_validators_common import (
    TotemBase,
    TransportKnetBase,
    TransportUdpBase,
)
from pcs_test.tools.assertions import assert_report_item_list_equal


class UpdateTotem(TotemBase, TestCase):
    def call_function(self, options):
        return config_validators.update_totem(options)

    def test_empty_values_allowed(self):
        assert_report_item_list_equal(
            self.call_function({name: "" for name in self.allowed_options}),
            [],
        )


class UpdateTransportKnet(TransportKnetBase, TestCase):
    def call_function(
        self,
        generic_options,
        compression_options,
        crypto_options,
        current_crypto_options=None,
    ):
        return config_validators.update_transport_knet(
            generic_options,
            compression_options,
            crypto_options,
            current_crypto_options=(
                {} if current_crypto_options is None else current_crypto_options
            ),
        )

    def test_empty_values_allowed(self):
        assert_report_item_list_equal(
            self.call_function(
                {
                    "ip_version": "",
                    "knet_pmtud_interval": "",
                    "link_mode": "",
                },
                {
                    "level": "",
                    "model": "",
                    "threshold": "",
                },
                {
                    "cipher": "",
                    "hash": "",
                    "model": "",
                },
            ),
            [],
        )

    def test_crypto_config_enabled_set_to_disabled(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "none", "hash": "none"},
                {"cipher": "aes256", "hash": "sha256"},
            ),
            [],
        )

    def test_crypto_config_enabled_set_to_default(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "", "hash": ""},
                {"cipher": "aes256", "hash": "sha256"},
            ),
            [],
        )

    def test_crypto_config_enabled_default_hash(self):
        assert_report_item_list_equal(
            self.call_function(
                {}, {}, {"hash": ""}, {"cipher": "aes256", "hash": "sha256"}
            ),
            [self.fixture_error_prerequisite],
        )

    def test_crypto_config_enabled_disabled_hash(self):
        assert_report_item_list_equal(
            self.call_function(
                {}, {}, {"hash": "none"}, {"cipher": "aes256", "hash": "sha256"}
            ),
            [self.fixture_error_prerequisite],
        )

    def test_crypto_config_enabled_changed_hash(self):
        assert_report_item_list_equal(
            self.call_function(
                {}, {}, {"hash": "md5"}, {"cipher": "aes256", "hash": "sha256"}
            ),
            [],
        )

    def test_crypto_config_enabled_changed_cipher(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "aes128"},
                {"cipher": "aes256", "hash": "sha256"},
            ),
            [],
        )

    def test_crypto_config_hash_enabled_enable_cipher(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "aes128"},
                {"hash": "sha256"},
            ),
            [],
        )

    def test_crypto_config_hash_enabled_enable_cipher_disable_hash(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "aes128", "hash": "none"},
                {"hash": "sha256"},
            ),
            [self.fixture_error_prerequisite],
        )

    def test_crypto_config_hash_enabled_enable_cipher_default_hash(self):
        assert_report_item_list_equal(
            self.call_function(
                {},
                {},
                {"cipher": "aes128", "hash": ""},
                {"hash": "sha256"},
            ),
            [self.fixture_error_prerequisite],
        )


class UpdateTransportUdp(TransportUdpBase, TestCase):
    def call_function(
        self, generic_options, compression_options, crypto_options
    ):
        return config_validators.update_transport_udp(
            generic_options, compression_options, crypto_options
        )

    def test_empty_values_allowed(self):
        assert_report_item_list_equal(
            self.call_function({"ip_version": "", "netmtu": ""}, {}, {}),
            [],
        )



import base64
import io
import re
from collections import namedtuple
from urllib.parse import urlencode

# We should ignore SIGPIPE when using pycurl.NOSIGNAL - see the libcurl tutorial
# for more info.
try:
    import signal

    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
except ImportError:
    pass

from pcs import settings
from pcs.common import pcs_pycurl as pycurl
from pcs.common.host import Destination


def _find_value_for_possible_keys(value_dict, possible_key_list):
    for key in possible_key_list:
        if key in value_dict:
            return value_dict[key]
    return None


class HostNotFound(Exception):
    def __init__(self, name):
        super().__init__()
        self.name = name


class NodeTargetFactory:
    def __init__(self, known_hosts):
        self._known_hosts = known_hosts

    def get_target(self, host_name):
        known_host = self._known_hosts.get(host_name)
        if known_host is None:
            raise HostNotFound(host_name)
        return RequestTarget.from_known_host(known_host)

    def get_target_from_hostname(self, hostname):
        try:
            return self.get_target(hostname)
        except HostNotFound:
            return RequestTarget(hostname)


class RequestData(
    namedtuple("RequestData", ["action", "structured_data", "data"])
):
    """
    This class represents action and data associated with action which will be
    send in request
    """

    def __new__(cls, action, structured_data=(), data=None):
        """
        string action -- action to perform
        list structured_data -- list of tuples, data to send with specified
            action
        string data -- raw data to send in request's body
        """
        return super(RequestData, cls).__new__(
            cls,
            action,
            data if data else structured_data,
            data if data else urlencode(structured_data),
        )


class RequestTarget(
    namedtuple("RequestTarget", ["label", "token", "dest_list"])
):
    """
    This class represents target (host) for request to be performed on
    """

    def __new__(cls, label, token=None, dest_list=()):
        if not dest_list:
            dest_list = [Destination(label, settings.pcsd_default_port)]
        return super(RequestTarget, cls).__new__(
            cls,
            label,
            token=token,
            dest_list=list(dest_list),
        )

    @classmethod
    def from_known_host(cls, known_host):
        return cls(
            known_host.name,
            token=known_host.token,
            dest_list=known_host.dest_list,
        )

    @property
    def first_addr(self):
        # __new__ ensures there is always at least one item in self.dest_list
        return self.dest_list[0].addr


class Request:
    """
    This class represents request. With usage of RequestTarget it provides
    interface for getting next available host to make request on.
    """

    def __init__(self, request_target, request_data):
        """
        RequestTarget request_target
        RequestData request_data
        """
        self._target = request_target
        self._data = request_data
        self._current_dest_iterator = iter(self._target.dest_list)
        self._current_dest = None
        self.next_dest()

    def next_dest(self):
        """
        Move to the next available host connection. Raises StopIteration when
        there is no connection to use.
        """
        self._current_dest = next(self._current_dest_iterator)

    @property
    def url(self):
        """
        URL representing request using current host.
        """
        addr = self.dest.addr
        port = self.dest.port
        return "https://{host}:{port}/{request}".format(
            host="[{0}]".format(addr) if ":" in addr else addr,
            port=(port if port else settings.pcsd_default_port),
            request=self._data.action,
        )

    @property
    def dest(self):
        return self._current_dest

    @property
    def host_label(self):
        return self._target.label

    @property
    def target(self):
        return self._target

    @property
    def data(self):
        return self._data.data

    @property
    def action(self):
        return self._data.action

    @property
    def cookies(self):
        cookies = {}
        if self._target.token:
            cookies["token"] = self._target.token
        return cookies

    def __repr__(self):
        return str("Request({0}, {1})").format(self._target, self._data)


class Response:
    """
    This class represents response for request which is available as instance
    property.
    """

    def __init__(self, handle, was_connected, errno=None, error_msg=None):
        self._handle = handle
        self._was_connected = was_connected
        self._errno = errno
        self._error_msg = error_msg
        self._data = None
        self._debug = None

    @classmethod
    def connection_successful(cls, handle):
        """
        Returns Response instance that is marked as successfully connected.

        pycurl.Curl handle -- curl easy handle, which connection was successful
        """
        return cls(handle, True)

    @classmethod
    def connection_failure(cls, handle, errno, error_msg):
        """
        Returns Response instance that is marked as not successfully connected.

        pycurl.Curl handle -- curl easy handle, which was not connected
        int errno -- error number
        string error_msg -- text description of error
        """
        return cls(handle, False, errno, error_msg)

    @property
    def request(self):
        return self._handle.request_obj

    @property
    def handle(self):
        return self._handle

    @property
    def was_connected(self):
        return self._was_connected

    @property
    def errno(self):
        return self._errno

    @property
    def error_msg(self):
        return self._error_msg

    @property
    def data(self):
        if self._data is None:
            self._data = self._handle.output_buffer.getvalue().decode("utf-8")
        return self._data

    @property
    def debug(self):
        if self._debug is None:
            self._debug = self._handle.debug_buffer.getvalue().decode("utf-8")
        return self._debug

    @property
    def response_code(self):
        if not self.was_connected:
            return None
        return self._handle.getinfo(pycurl.RESPONSE_CODE)

    def __repr__(self):
        return str(
            "Response({0} data='{1}' was_connected={2}) errno='{3}'"
            " error_msg='{4}' response_code='{5}')"
        ).format(
            self.request,
            self.data,
            self.was_connected,
            self.errno,
            self.error_msg,
            self.response_code,
        )


class NodeCommunicatorFactory:
    def __init__(self, communicator_logger, user, groups, request_timeout):
        self._logger = communicator_logger
        self._user = user
        self._groups = groups
        self._request_timeout = request_timeout

    def get_communicator(self, request_timeout=None):
        return self.get_simple_communicator(request_timeout=request_timeout)

    def get_simple_communicator(self, request_timeout=None):
        timeout = request_timeout if request_timeout else self._request_timeout
        return Communicator(
            self._logger, self._user, self._groups, request_timeout=timeout
        )

    def get_multiaddress_communicator(self, request_timeout=None):
        timeout = request_timeout if request_timeout else self._request_timeout
        return MultiaddressCommunicator(
            self._logger, self._user, self._groups, request_timeout=timeout
        )


class Communicator:
    """
    This class provides simple interface for making parallel requests.
    The instances of this class are not thread-safe! It is intended to use it
    only in a single thread. Use an unique instance for each thread.
    """

    curl_multi_select_timeout_default = 0.8  # in seconds

    def __init__(self, communicator_logger, user, groups, request_timeout=None):
        self._logger = communicator_logger
        self._auth_cookies = _get_auth_cookies(user, groups)
        self._request_timeout = (
            request_timeout
            if request_timeout is not None
            else settings.default_request_timeout
        )
        self._multi_handle = pycurl.CurlMulti()
        self._is_running = False
        # This is used just for storing references of curl easy handles.
        # We need to have references for all the handles, so they don't be
        # cleaned up by the garbage collector.
        self._easy_handle_list = []

    def add_requests(self, request_list):
        """
        Add requests to queue to be processed. It is possible to call this
        method before getting generator using start_loop method and also during
        getting responses from generator.  Requests are not performed after
        calling this method, but only when generator returned by start_loop
        method is in progress (returned at least one response and not raised
        StopIteration exception).

        list request_list -- Request objects to add to the queue
        """
        for request in request_list:
            handle = _create_request_handle(
                request,
                self._auth_cookies,
                self._request_timeout,
            )
            self._easy_handle_list.append(handle)
            self._multi_handle.add_handle(handle)
            if self._is_running:
                self._logger.log_request_start(request)

    def start_loop(self):
        """
        Returns generator. When generator is invoked, all requests in queue
        (added by method add_requests) will be invoked in parallel, and
        generator will then return responses for these requests. It is possible
        to add new request to the queue while the generator is in progress.
        Generator will stop (raise StopIteration) after all requests (also those
        added after creation of generator) are processed.

        WARNING: do not use multiple instances of generator (of one
        Communicator instance) when there is one which didn't finish
        (raised StopIteration). It will cause AssertionError.

        USAGE:
        com = Communicator(...)
        com.add_requests([
            Request(...), ...
        ])
        for response in communicator.start_loop():
            # do something with response
            # if needed, add some new requests to the queue
            com.add_requests([Request(...)])
        """
        if self._is_running:
            raise AssertionError("Method start_loop already running")
        self._is_running = True
        for handle in self._easy_handle_list:
            self._logger.log_request_start(handle.request_obj)

        finished_count = 0
        while finished_count < len(self._easy_handle_list):
            self.__multi_perform()
            self.__wait_for_multi_handle()
            response_list = self.__get_all_ready_responses()
            for response in response_list:
                # free up memory for next usage of this Communicator instance
                self._multi_handle.remove_handle(response.handle)
                self._logger.log_response(response)
                yield response
                # if something was added to the queue in the meantime, run it
                # immediately, so we don't need to wait until all responses will
                # be processed
                self.__multi_perform()
            finished_count += len(response_list)
        self._easy_handle_list = []
        self._is_running = False

    def __get_all_ready_responses(self):
        response_list = []
        repeat = True
        while repeat:
            num_queued, ok_list, err_list = self._multi_handle.info_read()
            response_list.extend(
                [Response.connection_successful(handle) for handle in ok_list]
                + [
                    Response.connection_failure(handle, errno, error_msg)
                    for handle, errno, error_msg in err_list
                ]
            )
            repeat = num_queued > 0
        return response_list

    def __multi_perform(self):
        # run all internal operation required by libcurl
        status, num_to_process = self._multi_handle.perform()
        # if perform returns E_CALL_MULTI_PERFORM it requires to call perform
        # once again right away
        while status == pycurl.E_CALL_MULTI_PERFORM:
            status, num_to_process = self._multi_handle.perform()
        return num_to_process

    def __wait_for_multi_handle(self):
        # try to wait until there is something to do for us
        need_to_wait = True
        while need_to_wait:
            timeout = self._multi_handle.timeout()
            if timeout == 0:
                # if timeout == 0 then there is something to precess already
                return
            timeout = (
                timeout / 1000.0
                if timeout > 0
                # curl don't have timeout set, so we can use our default
                else self.curl_multi_select_timeout_default
            )
            # when value returned from select is -1, it timed out, so we can
            # wait
            need_to_wait = self._multi_handle.select(timeout) == -1


class MultiaddressCommunicator(Communicator):
    """
    Class with same interface as Communicator. In difference with Communicator,
    it takes advantage of multiple hosts in RequestTarget. So if it is not
    possible to connect to target using first hostname, it will use next one
    until connection will be successful or there is no host left.
    """

    def start_loop(self):
        for response in super().start_loop():
            if response.was_connected:
                yield response
                continue
            try:
                previous_dest = response.request.dest
                response.request.next_dest()
                self._logger.log_retry(response, previous_dest)
                self.add_requests([response.request])
            except StopIteration:
                self._logger.log_no_more_addresses(response)
                yield response


class CommunicatorLoggerInterface:
    def log_request_start(self, request):
        raise NotImplementedError()

    def log_response(self, response):
        raise NotImplementedError()

    def log_retry(self, response, previous_dest):
        raise NotImplementedError()

    def log_no_more_addresses(self, response):
        raise NotImplementedError()


def _get_auth_cookies(user, group_list):
    """
    Returns input parameters in a dictionary which is prepared to be converted
    to cookie string.

    string user -- CIB user
    string group_list -- CIB user groups
    """
    # Let's be safe about characters in variables (they can come from env)
    # and do base64. We cannot do it for CIB_user however to be backward
    # compatible so we at least remove disallowed characters.
    cookies = {}
    if user:
        cookies["CIB_user"] = re.sub(r"[^!-~]", "", user).replace(";", "")
    if group_list:
        # cookies require string but base64encode returns bytes, so decode it...
        cookies["CIB_user_groups"] = base64.b64encode(
            # python3 requires the value to be bytes not str
            " ".join(group_list).encode("utf-8")
        ).decode("utf-8")
    return cookies


def _create_request_handle(request, cookies, timeout):
    """
    Returns Curl object (easy handle) which is set up witc specified parameters.

    Request request -- request specification
    dict cookies -- cookies to add to request
    int timeout -- request timeout
    """
    # it is not possible to take this callback out of this function, because of
    # curl API
    def __debug_callback(data_type, debug_data):
        # pylint: disable=no-member
        prefixes = {
            pycurl.DEBUG_TEXT: b"* ",
            pycurl.DEBUG_HEADER_IN: b"< ",
            pycurl.DEBUG_HEADER_OUT: b"> ",
            pycurl.DEBUG_DATA_IN: b"<< ",
            pycurl.DEBUG_DATA_OUT: b">> ",
        }
        if data_type in prefixes:
            debug_output.write(prefixes[data_type])
            debug_output.write(debug_data)
            if not debug_data.endswith(b"\n"):
                debug_output.write(b"\n")

    output = io.BytesIO()
    debug_output = io.BytesIO()
    cookies.update(request.cookies)
    handle = pycurl.Curl()
    handle.setopt(pycurl.PROTOCOLS, pycurl.PROTO_HTTPS)
    handle.setopt(pycurl.TIMEOUT, timeout)
    handle.setopt(pycurl.URL, request.url.encode("utf-8"))
    handle.setopt(pycurl.WRITEFUNCTION, output.write)
    handle.setopt(pycurl.VERBOSE, 1)
    handle.setopt(pycurl.DEBUGFUNCTION, __debug_callback)
    handle.setopt(pycurl.SSL_VERIFYHOST, 0)
    handle.setopt(pycurl.SSL_VERIFYPEER, 0)
    handle.setopt(pycurl.NOSIGNAL, 1)  # required for multi-threading
    handle.setopt(pycurl.HTTPHEADER, ["Expect: "])
    if cookies:
        handle.setopt(pycurl.COOKIE, _dict_to_cookies(cookies).encode("utf-8"))
    if request.data:
        handle.setopt(pycurl.COPYPOSTFIELDS, request.data.encode("utf-8"))
    # add reference for request object and output buffers to handle, so later
    # we don't need to match these objects when they are returned from
    # pycurl after they've been processed
    # similar usage is in pycurl example:
    # https://github.com/pycurl/pycurl/blob/REL_7_19_0_3/examples/retriever-multi.py
    handle.request_obj = request
    handle.output_buffer = output
    handle.debug_buffer = debug_output
    return handle


def _dict_to_cookies(cookies_dict):
    return ";".join(
        [
            "{0}={1}".format(key, value)
            for key, value in sorted(cookies_dict.items())
        ]
    )



from pcs.common.pacemaker.resource.operations import (
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
)

OPERATION_ATTRIBUTES = [
    "id",
    "description",
    "enabled",
    "interval",
    "interval-origin",
    "name",
    "on-fail",
    "record-pending",
    "role",
    "start-delay",
    "timeout",
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
]



from unittest import (
    TestCase,
    mock,
)

import pcs.lib.sbd as lib_sbd
from pcs import settings
from pcs.common import reports
from pcs.common.reports import ReportItemSeverity as Severities
from pcs.common.reports import codes as report_codes
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor
from pcs_test.tools.misc import outdent


class TestException(Exception):
    pass


class EvenNumberOfNodesAndNoQdevice(TestCase):
    # pylint: disable=protected-access
    def setUp(self):
        self.mock_corosync_conf = mock.MagicMock(spec_set=CorosyncConfigFacade)

    def _set_ret_vals(self, nodes, qdevice):
        self.mock_corosync_conf.get_nodes.return_value = nodes
        self.mock_corosync_conf.has_quorum_device.return_value = qdevice

    def test_even_num_no_qdevice(self):
        self._set_ret_vals([1, 2], False)
        self.assertTrue(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf
            )
        )

    def test_even_num_qdevice(self):
        self._set_ret_vals([1, 2], True)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf
            )
        )

    def test_odd_num_no_qdevice(self):
        self._set_ret_vals([1, 2, 3], False)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf
            )
        )

    def test_odd_num_qdevice(self):
        self._set_ret_vals([1, 2, 3], True)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf
            )
        )

    def test_even_num_no_qdevice_plus_one(self):
        self._set_ret_vals([1, 2], False)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf, 1
            )
        )

    def test_even_num_qdevice_plus_one(self):
        self._set_ret_vals([1, 2], True)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf, 1
            )
        )

    def test_odd_num_no_qdevice_plus_one(self):
        self._set_ret_vals([1, 2, 3], False)
        self.assertTrue(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf, 1
            )
        )

    def test_odd_num_qdevice_plus_one(self):
        self._set_ret_vals([1, 2, 3], True)
        self.assertFalse(
            lib_sbd._even_number_of_nodes_and_no_qdevice(
                self.mock_corosync_conf, 1
            )
        )


@mock.patch("pcs.lib.sbd._is_device_set_local")
@mock.patch("pcs.lib.sbd.is_sbd_enabled")
@mock.patch("pcs.lib.sbd.is_sbd_installed")
@mock.patch("pcs.lib.sbd._even_number_of_nodes_and_no_qdevice")
class IsAutoTieBreakerNeededTest(TestCase):
    def setUp(self):
        self.runner = "runner"
        self.corosync_conf_facade = "facade"
        self.node_num_modifier = 1
        self.mock_nodes_and_qdevice = None
        self.mock_sbd_installed = None
        self.mock_sbd_enabled = None
        self.mock_device_set = None

    def set_mocks(
        self,
        mock_nodes_and_qdevice,
        mock_sbd_installed,
        mock_sbd_enabled,
        mock_device_set,
    ):
        self.mock_nodes_and_qdevice = mock_nodes_and_qdevice
        self.mock_sbd_installed = mock_sbd_installed
        self.mock_sbd_enabled = mock_sbd_enabled
        self.mock_device_set = mock_device_set

    def common_test(
        self, nodes_and_qdevice, sbd_installed, sbd_enabled, device_set, result
    ):
        self.mock_nodes_and_qdevice.return_value = nodes_and_qdevice
        self.mock_sbd_installed.return_value = sbd_installed
        self.mock_sbd_enabled.return_value = sbd_enabled
        self.mock_device_set.return_value = device_set
        self.assertEqual(
            lib_sbd.is_auto_tie_breaker_needed(
                self.runner, self.corosync_conf_facade, self.node_num_modifier
            ),
            result,
        )
        self.mock_nodes_and_qdevice.assert_called_once_with(
            self.corosync_conf_facade, self.node_num_modifier
        )

    def test_device_set(
        self,
        mock_nodes_and_qdevice,
        mock_sbd_installed,
        mock_sbd_enabled,
        mock_device_set,
    ):
        self.set_mocks(
            mock_nodes_and_qdevice,
            mock_sbd_installed,
            mock_sbd_enabled,
            mock_device_set,
        )
        self.common_test(True, True, True, True, False)

    def test_no_device(
        self,
        mock_nodes_and_qdevice,
        mock_sbd_installed,
        mock_sbd_enabled,
        mock_device_set,
    ):
        self.set_mocks(
            mock_nodes_and_qdevice,
            mock_sbd_installed,
            mock_sbd_enabled,
            mock_device_set,
        )
        self.common_test(True, True, True, False, True)

    def test_no_device_with_qdevice(
        self,
        mock_nodes_and_qdevice,
        mock_sbd_installed,
        mock_sbd_enabled,
        mock_device_set,
    ):
        self.set_mocks(
            mock_nodes_and_qdevice,
            mock_sbd_installed,
            mock_sbd_enabled,
            mock_device_set,
        )
        self.common_test(False, True, True, False, False)

    def test_sbd_disabled(
        self,
        mock_nodes_and_qdevice,
        mock_sbd_installed,
        mock_sbd_enabled,
        mock_device_set,
    ):
        self.set_mocks(
            mock_nodes_and_qdevice,
            mock_sbd_installed,
            mock_sbd_enabled,
            mock_device_set,
        )
        self.common_test(True, True, False, False, False)


@mock.patch("pcs.lib.sbd.is_auto_tie_breaker_needed")
class AtbHasToBeEnabledTest(TestCase):
    def setUp(self):
        self.mock_service_manager = "service manager mock"
        self.mock_conf = mock.MagicMock(spec_set=CorosyncConfigFacade)

    def test_atb_needed_is_enabled(self, mock_is_needed):
        mock_is_needed.return_value = True
        self.mock_conf.is_enabled_auto_tie_breaker.return_value = True
        self.assertFalse(
            lib_sbd.atb_has_to_be_enabled(
                self.mock_service_manager, self.mock_conf, 1
            )
        )
        self.mock_conf.is_enabled_auto_tie_breaker.assert_called_once_with()
        mock_is_needed.assert_not_called()

    def test_atb_needed_is_disabled(self, mock_is_needed):
        mock_is_needed.return_value = True
        self.mock_conf.is_enabled_auto_tie_breaker.return_value = False
        self.assertTrue(
            lib_sbd.atb_has_to_be_enabled(
                self.mock_service_manager, self.mock_conf, -1
            )
        )
        self.mock_conf.is_enabled_auto_tie_breaker.assert_called_once_with()
        mock_is_needed.assert_called_once_with(
            self.mock_service_manager, self.mock_conf, -1
        )

    def test_atb_not_needed_is_enabled(self, mock_is_needed):
        mock_is_needed.return_value = False
        self.mock_conf.is_enabled_auto_tie_breaker.return_value = True
        self.assertFalse(
            lib_sbd.atb_has_to_be_enabled(
                self.mock_service_manager, self.mock_conf, 2
            )
        )
        self.mock_conf.is_enabled_auto_tie_breaker.assert_called_once_with()
        mock_is_needed.assert_not_called()

    def test_atb_not_needed_is_disabled(self, mock_is_needed):
        mock_is_needed.return_value = False
        self.mock_conf.is_enabled_auto_tie_breaker.return_value = False
        self.assertFalse(
            lib_sbd.atb_has_to_be_enabled(
                self.mock_service_manager, self.mock_conf, -2
            )
        )
        self.mock_conf.is_enabled_auto_tie_breaker.assert_called_once_with()
        mock_is_needed.assert_called_once_with(
            self.mock_service_manager, self.mock_conf, -2
        )


@mock.patch("pcs.lib.sbd.is_systemd")
class GetSbdServiceNameTest(TestCase):
    def setUp(self):
        self.service_manager = mock.MagicMock()

    def test_systemctl(self, mock_is_systemd):
        mock_is_systemd.return_value = True
        self.assertEqual(
            "sbd", lib_sbd.get_sbd_service_name(self.service_manager)
        )
        mock_is_systemd.assert_called_once_with(self.service_manager)

    def test_not_systemctl(self, mock_is_systemd):
        mock_is_systemd.return_value = False
        self.assertEqual(
            "sbd_helper", lib_sbd.get_sbd_service_name(self.service_manager)
        )
        mock_is_systemd.assert_called_once_with(self.service_manager)


@mock.patch("pcs.lib.sbd.get_sbd_service_name")
class IsSbdEnabledTest(TestCase):
    def test_success(self, mock_sbd_name):
        mock_obj = mock.MagicMock()
        mock_obj.is_enabled.return_value = True
        service = "sbd"
        mock_sbd_name.return_value = service
        self.assertTrue(lib_sbd.is_sbd_enabled(mock_obj))
        mock_obj.is_enabled.assert_called_once_with(service)
        mock_sbd_name.assert_called_once_with(mock_obj)


@mock.patch("pcs.lib.sbd.get_sbd_service_name")
class IsSbdInstalledTest(TestCase):
    def setUp(self):
        self.service = "sbd"

    def test_installed(self, mock_sbd_name):
        mock_obj = mock.MagicMock()
        mock_obj.is_installed.return_value = True
        mock_sbd_name.return_value = self.service
        self.assertTrue(lib_sbd.is_sbd_installed(mock_obj))
        mock_obj.is_installed.assert_called_once_with(self.service)
        mock_sbd_name.assert_called_once_with(mock_obj)

    def test_not_installed(self, mock_sbd_name):
        mock_obj = mock.MagicMock()
        mock_obj.is_installed.return_value = False
        mock_sbd_name.return_value = self.service
        self.assertFalse(lib_sbd.is_sbd_installed(mock_obj))
        mock_obj.is_installed.assert_called_once_with(self.service)
        mock_sbd_name.assert_called_once_with(mock_obj)


class InitializeBlockDeviceTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock()
        self.mock_rep = MockLibraryReportProcessor()

    def test_success(self):
        device_list = ["/dev/sdb", "/dev/vda"]
        option_dict = {
            "watchdog-timeout": "10",  # -1
            "loop-timeout": "1",  # -3
        }
        self.mock_runner.run.return_value = "", "", 0
        lib_sbd.initialize_block_devices(
            self.mock_rep, self.mock_runner, device_list, option_dict
        )
        cmd = [
            settings.sbd_binary,
            "-d",
            "/dev/sdb",
            "-d",
            "/dev/vda",
            "-3",
            "1",
            "-1",
            "10",
            "create",
        ]
        self.mock_runner.run.assert_called_once_with(cmd)
        assert_report_item_list_equal(
            self.mock_rep.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.SBD_DEVICE_INITIALIZATION_STARTED,
                    {"device_list": device_list},
                ),
                (
                    Severities.INFO,
                    report_codes.SBD_DEVICE_INITIALIZATION_SUCCESS,
                    {"device_list": device_list},
                ),
            ],
        )

    def test_failed(self):
        device_list = ["/dev/sdb", "/dev/vda"]
        option_dict = {
            "watchdog-timeout": "10",  # -1
            "loop-timeout": "1",  # -3
        }
        error_msg = "error"
        self.mock_runner.run.return_value = "", error_msg, 1
        assert_raise_library_error(
            lambda: lib_sbd.initialize_block_devices(
                self.mock_rep, self.mock_runner, device_list, option_dict
            ),
            (
                Severities.ERROR,
                report_codes.SBD_DEVICE_INITIALIZATION_ERROR,
                {
                    "device_list": device_list,
                    "reason": error_msg,
                },
            ),
        )
        cmd = [
            settings.sbd_binary,
            "-d",
            "/dev/sdb",
            "-d",
            "/dev/vda",
            "-3",
            "1",
            "-1",
            "10",
            "create",
        ]
        self.mock_runner.run.assert_called_once_with(cmd)
        assert_report_item_list_equal(
            self.mock_rep.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.SBD_DEVICE_INITIALIZATION_STARTED,
                    {"device_list": device_list},
                )
            ],
        )


@mock.patch("os.path.exists")
@mock.patch("pcs.lib.sbd.get_local_sbd_config")
class GetLocalSbdDeviceListTest(TestCase):
    def test_device_not_defined(self, mock_sbd_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_sbd_config.return_value = outdent(
            """
            SBD_WATCHDOG=/dev/watchdog
            SBD_WATCHDOG_TIMEOUT=10
        """
        )
        self.assertEqual([], lib_sbd.get_local_sbd_device_list())
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        mock_sbd_config.assert_called_once_with()

    def test_no_device(self, mock_sbd_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_sbd_config.return_value = outdent(
            """
            SBD_WATCHDOG=/dev/watchdog
            SBD_WATCHDOG_TIMEOUT=10
            SBD_DEVICE=""
        """
        )
        self.assertEqual([], lib_sbd.get_local_sbd_device_list())
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        mock_sbd_config.assert_called_once_with()

    def test_one_device(self, mock_sbd_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_sbd_config.return_value = outdent(
            """
            SBD_WATCHDOG=/dev/watchdog
            SBD_WATCHDOG_TIMEOUT=10
            SBD_DEVICE="/dev/vda"
        """
        )
        self.assertEqual(["/dev/vda"], lib_sbd.get_local_sbd_device_list())
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        mock_sbd_config.assert_called_once_with()

    def test_multiple_devices(self, mock_sbd_config, mock_config_exists):
        mock_config_exists.return_value = True
        mock_sbd_config.return_value = outdent(
            """
            SBD_WATCHDOG=/dev/watchdog
            SBD_WATCHDOG_TIMEOUT=10
            SBD_DEVICE="/dev/vda;/dev/sda"
        """
        )
        self.assertEqual(
            ["/dev/vda", "/dev/sda"], lib_sbd.get_local_sbd_device_list()
        )
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        mock_sbd_config.assert_called_once_with()

    def test_config_does_not_exist(self, mock_sbd_config, mock_config_exists):
        mock_config_exists.return_value = False
        self.assertEqual([], lib_sbd.get_local_sbd_device_list())
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        self.assertEqual(0, mock_sbd_config.call_count)

    def test_config_read_error(self, mock_sbd_config, mock_config_exists):
        # pylint: disable=no-self-use
        mock_config_exists.return_value = True
        node = "local node"
        error = "error string"
        mock_sbd_config.side_effect = LibraryError(
            reports.item.ReportItem.error(
                reports.messages.UnableToGetSbdConfig(node, error)
            )
        )
        assert_raise_library_error(
            lib_sbd.get_local_sbd_device_list,
            (
                Severities.ERROR,
                report_codes.UNABLE_TO_GET_SBD_CONFIG,
                {
                    "node": node,
                    "reason": error,
                },
            ),
        )
        mock_config_exists.assert_called_once_with(settings.sbd_config)
        mock_sbd_config.assert_called_once_with()


@mock.patch("pcs.lib.sbd.get_local_sbd_device_list")
class IsDeviceSetLocalTest(TestCase):
    # pylint: disable=protected-access
    def test_no_device(self, mock_device_list):
        mock_device_list.return_value = []
        self.assertFalse(lib_sbd._is_device_set_local())
        mock_device_list.assert_called_once_with()

    def test_one_device(self, mock_device_list):
        mock_device_list.return_value = ["device1"]
        self.assertTrue(lib_sbd._is_device_set_local())
        mock_device_list.assert_called_once_with()

    def test_multiple_devices(self, mock_device_list):
        mock_device_list.return_value = ["device1", "device2"]
        self.assertTrue(lib_sbd._is_device_set_local())
        mock_device_list.assert_called_once_with()


class GetDeviceMessagesInfoTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self):
        output = "out"
        device = "device"
        self.mock_runner.run.return_value = output, "", 0
        self.assertEqual(
            output, lib_sbd.get_device_messages_info(self.mock_runner, device)
        )
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", device, "list"]
        )

    def test_failed(self):
        output = "out"
        device = "device"
        self.mock_runner.run.return_value = output, "", 1
        assert_raise_library_error(
            lambda: lib_sbd.get_device_messages_info(self.mock_runner, device),
            (
                Severities.ERROR,
                report_codes.SBD_DEVICE_LIST_ERROR,
                {
                    "device": device,
                    "reason": output,
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", device, "list"]
        )


class GetDeviceSbdHeaderDumpTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self):
        output = "out"
        device = "device"
        self.mock_runner.run.return_value = output, "", 0
        self.assertEqual(
            output, lib_sbd.get_device_sbd_header_dump(self.mock_runner, device)
        )
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", device, "dump"]
        )

    def test_failed(self):
        output = "out"
        device = "device"
        self.mock_runner.run.return_value = output, "", 1
        assert_raise_library_error(
            lambda: lib_sbd.get_device_sbd_header_dump(
                self.mock_runner, device
            ),
            (
                Severities.ERROR,
                report_codes.SBD_DEVICE_DUMP_ERROR,
                {
                    "device": device,
                    "reason": output,
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", device, "dump"]
        )


class SetMessageTest(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def test_success(self):
        self.mock_runner.run.return_value = "", "", 0
        lib_sbd.set_message(self.mock_runner, "device", "node", "test")
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", "device", "message", "node", "test"]
        )

    def test_failure(self):
        error = "error"
        self.mock_runner.run.return_value = "", error, 1
        assert_raise_library_error(
            lambda: lib_sbd.set_message(
                self.mock_runner, "device", "node", "test"
            ),
            (
                Severities.ERROR,
                report_codes.SBD_DEVICE_MESSAGE_ERROR,
                {
                    "device": "device",
                    "node": "node",
                    "sbd_message": "test",
                    "reason": error,
                },
            ),
        )
        self.mock_runner.run.assert_called_once_with(
            [settings.sbd_binary, "-d", "device", "message", "node", "test"]
        )


class ValidateDeviceDictTest(TestCase):
    def test_all_ok(self):
        device_dict = {
            "node1": ["/dev1", "/dev2"],
            "node2": ["/dev1"],
        }
        self.assertEqual([], lib_sbd.validate_nodes_devices(device_dict))

    def test_some_not_ok(self):
        # pylint: disable=no-self-use
        too_many_devices = [
            "/dev" + str(i) for i in range(settings.sbd_max_device_num)
        ] + ["dev/sda2"]
        device_dict = {
            "node1": [],
            "node2": too_many_devices,
            "node3": ["/dev/vda"],
            "node4": ["/dev/vda1", "../dev/sda2"],
        }
        assert_report_item_list_equal(
            lib_sbd.validate_nodes_devices(device_dict),
            [
                (
                    Severities.ERROR,
                    report_codes.SBD_NO_DEVICE_FOR_NODE,
                    {
                        "node": "node1",
                        "sbd_enabled_in_cluster": False,
                    },
                ),
                (
                    Severities.ERROR,
                    report_codes.SBD_TOO_MANY_DEVICES_FOR_NODE,
                    {
                        "node": "node2",
                        "device_list": too_many_devices,
                        "max_devices": settings.sbd_max_device_num,
                    },
                ),
                (
                    Severities.ERROR,
                    report_codes.SBD_DEVICE_PATH_NOT_ABSOLUTE,
                    {
                        "node": "node2",
                        "device": "dev/sda2",
                    },
                ),
                (
                    Severities.ERROR,
                    report_codes.SBD_DEVICE_PATH_NOT_ABSOLUTE,
                    {
                        "node": "node4",
                        "device": "../dev/sda2",
                    },
                ),
            ],
        )


class GetAvailableWatchdogs(TestCase):
    """
    Tested in pcs_test.tier0.lib.commands.sbd.test_watchdog_list
    """



from pcs import (
    config,
    usage,
)
from pcs.cli.common.routing import create_router

config_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.config(argv),
        "show": config.config_show,
        "backup": config.config_backup,
        "restore": config.config_restore,
        "checkpoint": create_router(
            {
                "list": config.config_checkpoint_list,
                "view": config.config_checkpoint_view,
                "restore": config.config_checkpoint_restore,
                "diff": config.config_checkpoint_diff,
            },
            ["config", "checkpoint"],
            default_cmd="list",
        ),
        # TODO remove, deprecated command
        "import-cman": config.config_import_cman,
        # TODO remove, deprecated command
        "export": create_router(
            {
                "pcs-commands": config.config_export_pcs_commands,
                "pcs-commands-verbose": lambda lib, argv, modifiers: (
                    config.config_export_pcs_commands(
                        lib, argv, modifiers, verbose=True
                    )
                ),
            },
            ["config", "export"],
        ),
    },
    ["config"],
    default_cmd="show",
)



from unittest import TestCase

from pcs.common.reports.constraints import ticket


class ConstraintPlainTest(TestCase):
    def test_prepare_report(self):
        self.assertEqual(
            "Master resourceA (id:some_id)",
            ticket.constraint_plain(
                {
                    "options": {
                        "rsc-role": "Master",
                        "rsc": "resourceA",
                        "id": "some_id",
                    }
                },
                with_id=True,
            ),
        )

    def test_prepare_report_without_role(self):
        self.assertEqual(
            "resourceA (id:some_id)",
            ticket.constraint_plain(
                {"options": {"rsc": "resourceA", "id": "some_id"}}, with_id=True
            ),
        )



import os
import ssl
from unittest import (
    TestCase,
    mock,
)

from pcs.daemon.ssl import (
    CertKeyPair,
    PcsdSSL,
    SSLCertKeyException,
)

from pcs_test.tools.misc import get_tmp_dir

SERVER_NAME = "pcsd-daemon"
SSL_OPTIONS = 0
SSL_CIPHERS = "DEFAULT:!RC4"


class SslFilesMixin:
    def setUp(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.ssl_dir = get_tmp_dir("tier0_daemon_ssl")
        self.cert_path = os.path.join(self.ssl_dir.name, "daemon.cert")
        self.key_path = os.path.join(self.ssl_dir.name, "daemon.key")

    def tearDown(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.ssl_dir.cleanup()

    def damage_ssl_files(self):
        with open(self.cert_path, "w") as cert:
            cert.write("bad content")
        with open(self.key_path, "w") as key:
            key.write("bad content")


class Pair(SslFilesMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pair = CertKeyPair(self.cert_path, self.key_path)

    def test_generated_key_check_success(self):
        self.pair.regenerate(SERVER_NAME)
        self.assertEqual(self.pair.check(), [])

    def test_error_if_files_with_bad_content(self):
        self.damage_ssl_files()
        errors = self.pair.check()
        self.assertEqual(len(errors), 1)
        self.assertRegex(
            errors[0],
            r"^SSL certificate does not match the key: "
            r"\[SSL\] PEM lib \(_ssl\.c:\d+\)",
        )

    @mock.patch("pcs.daemon.ssl.ssl.SSLContext.load_cert_chain")
    def test_error_if_short_key(self, mock_load_cert_chain):
        mock_load_cert_chain.side_effect = ssl.SSLError(
            # These are the real args of the exception.
            336245135,
            "[SSL: EE_KEY_TOO_SMALL] ee key too small (_ssl.c:3542)",
        )
        # 512 cannot be used as we would get an error from FIPS and 1024 is
        # long enough. So a mock must be used.
        self.pair.regenerate(SERVER_NAME, 1024)
        errors = self.pair.check()
        self.assertEqual(
            errors,
            [
                "SSL certificate does not match the key: "
                "[SSL: EE_KEY_TOO_SMALL] ee key too small (_ssl.c:3542)",
            ],
        )

    def test_error_if_cert_does_not_match_key(self):
        self.pair.regenerate(SERVER_NAME)
        with open(self.key_path, "rb") as key_file:
            key_content = key_file.read()
        self.pair.regenerate(SERVER_NAME)
        with open(self.key_path, "wb") as key_file:
            key_file.write(key_content)

        errors = self.pair.check()
        self.assertEqual(len(errors), 1)
        self.assertRegex(
            errors[0],
            r"SSL certificate does not match the key: "
            r"\[X509: KEY_VALUES_MISMATCH\] key values mismatch \(_ssl\.c:\d+\)",
        )


class PcsdSSLTest(SslFilesMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pcsd_ssl = PcsdSSL(
            SERVER_NAME, self.cert_path, self.key_path, SSL_OPTIONS, SSL_CIPHERS
        )

    def test_is_ok_when_ssl_files_does_not_exist(self):
        self.pcsd_ssl.guarantee_valid_certs()

    def test_raises_when_ssl_files_are_damaged(self):
        self.damage_ssl_files()
        with self.assertRaises(SSLCertKeyException) as ctx_manager:
            self.pcsd_ssl.guarantee_valid_certs()
        errors = ctx_manager.exception.args
        self.assertEqual(len(errors), 1)
        self.assertRegex(
            errors[0],
            r"SSL certificate does not match the key: "
            r"\[SSL\] PEM lib \(_ssl\.c:\d+\)",
        )

    def test_context_uses_given_options(self):
        self.pcsd_ssl.guarantee_valid_certs()
        ssl_context = self.pcsd_ssl.create_context()
        self.assertEqual(ssl_context.options, SSL_OPTIONS)



# pylint: disable=line-too-long
from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    is_minimum_pacemaker_version,
    is_pacemaker_21_without_20_compatibility,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

PCMK_2_0_3_PLUS = is_minimum_pacemaker_version(2, 0, 3)


class StonithWarningTest(TestCase, AssertPcsMixin):
    empty_cib = rc("cib-empty.xml")
    corosync_conf = rc("corosync.conf")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier0_statust_stonith_warning")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def fixture_stonith_action(self):
        self.assert_pcs_success(
            (
                "stonith create Sa fence_apc ip=i username=u action=reboot "
                "--force"
            ).split(),
            stdout_start="Warning: stonith option 'action' is deprecated and should not be "
            "used, use 'pcmk_off_action', 'pcmk_reboot_action' instead\n",
        )

    def fixture_stonith_cycle(self):
        self.assert_pcs_success_ignore_output(
            (
                "stonith",
                "create",
                "Sc",
                "fence_ipmilan",
                "method=cycle",
                "--force",
            )
        )

    def fixture_resource(self):
        self.assert_pcs_success(
            (
                "resource create dummy ocf:pacemaker:Dummy action=reboot "
                "method=cycle --force"
            ).split(),
            "Warning: invalid resource options: 'action', 'method', allowed "
            "options are: 'envfile', 'fail_start_on', 'fake', 'op_sleep', "
            "'passwd', 'state', 'trace_file', 'trace_ra'\n",
        )

    def test_warning_stonith_action(self):
        self.fixture_stonith_action()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'

                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'

                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )

    def test_warning_stonith_method_cycle(self):
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )

    def test_stonith_warnings(self):
        self.fixture_stonith_action()
        self.fixture_stonith_cycle()
        self.fixture_resource()
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'Sa'
                    Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'Sc'

                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )

    def test_warn_when_no_stonith(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    No stonith devices and stonith-enabled is not false

                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    WARNINGS:
                    No stonith devices and stonith-enabled is not false

                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )

    def test_no_stonith_warning_when_stonith_in_group(self):
        self.assert_pcs_success("stonith create S fence_xvm --group G".split())
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_binary")
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99
                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99

                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )

    def test_disabled_stonith_does_not_care_about_missing_devices(self):
        self.assert_pcs_success("property set stonith-enabled=false".split())
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        if PCMK_2_0_3_PLUS:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99
                    Cluster Summary:
                """
                ),
            )
        else:
            self.assert_pcs_success(
                ["status"],
                stdout_start=dedent(
                    """\
                    Cluster name: test99
                    Stack: unknown
                    Current DC: NONE
                """
                ),
            )


class ResourceStonithStatusBase(AssertPcsMixin):
    # pylint: disable=too-many-public-methods
    command = None
    no_resources_msg = None
    all_resources_output = None
    active_resources_output = None
    active_resources_output_node = None
    node_output = None
    cib_file = rc("cib-tags.xml")
    corosync_conf = rc("corosync.conf")
    no_active_resources_msg = "No active resources\n"

    def setUp(self):
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_status_resource_stonith_status")
        write_file_to_tmpfile(self.cib_file, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        # pylint: disable=invalid-name
        self.temp_cib.close()

    def test_not_resource_or_tag_id(self):
        self.assert_pcs_fail(
            self.command + ["cx1"],
            stdout_full="Error: resource or tag id 'cx1' not found\n",
        )

    def test_nonexistent_id(self):
        self.assert_pcs_fail(
            self.command + ["nonexistent"],
            stdout_full="Error: resource or tag id 'nonexistent' not found\n",
        )

    def test_missing_node_value(self):
        self.assert_pcs_fail(
            self.command + ["node="],
            stdout_full="Error: missing value of 'node' option\n",
        )

    def test_missing_node_key(self):
        self.assert_pcs_fail(
            self.command + ["=node"],
            stdout_full="Error: missing key in '=node' option\n",
        )

    def test_more_node_options(self):
        self.assert_pcs_fail(
            self.command + ["node=rh-1", "node=rh-2"],
            stdout_full=(
                "Error: duplicate option 'node' with different values 'rh-1' "
                "and 'rh-2'\n"
            ),
        )

    def test_more_no_node_option(self):
        self.assert_pcs_fail(
            self.command + ["r1", "r2"],
            stdout_full="Error: missing value of 'r2' option\n",
        )

    def test_resource_id(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = "  * x1	(ocf:pacemaker:Dummy):	 Started rh-1\n"
        else:
            stdout_full = "  * x1	(ocf::pacemaker:Dummy):	 Started rh-1\n"
        self.assert_pcs_success(
            self.command + ["x1"],
            stdout_full=stdout_full,
        )

    def test_resource_id_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["x2", "--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_resource_id_with_node_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["x2", "node=rh-1", "--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_resource_id_with_node_started(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = "  * x1	(ocf:pacemaker:Dummy):	 Started rh-1\n"
        else:
            stdout_full = "  * x1	(ocf::pacemaker:Dummy):	 Started rh-1\n"
        self.assert_pcs_success(
            self.command + ["x1", "node=rh-1"],
            stdout_full=stdout_full,
        )

    def test_resource_id_with_node_stopped(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = "  * x2	(ocf:pacemaker:Dummy):	 Stopped\n"
        else:
            stdout_full = "  * x2	(ocf::pacemaker:Dummy):	 Stopped\n"
        self.assert_pcs_success(
            self.command + ["x2", "node=rh-1"],
            stdout_full=stdout_full,
        )

    def test_resource_id_with_node_without_status(self):
        self.assert_pcs_success(
            self.command + ["x1", "node=rh-2"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_resource_id_with_node_changed_arg_order(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = "  * x1	(ocf:pacemaker:Dummy):	 Started rh-1\n"
        else:
            stdout_full = "  * x1	(ocf::pacemaker:Dummy):	 Started rh-1\n"
        self.assert_pcs_success(
            self.command + ["node=rh-1", "x1"],
            stdout_full=stdout_full,
        )

    def test_stonith_id(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-1"],
            stdout_full="  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1\n",
        )

    def test_stonith_id_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-2", "--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_stonith_id_with_node_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-2", "node=rh-2", "--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_stonith_id_with_node_started(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-1", "node=rh-1"],
            stdout_full="  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1\n",
        )

    def test_stonith_id_with_node_stopped(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-2", "node=rh-2"],
            stdout_full="  * fence-rh-2	(stonith:fence_xvm):	 Stopped\n",
        )

    def test_stonith_id_with_node_without_status(self):
        self.assert_pcs_success(
            self.command + ["fence-rh-1", "node=rh-2"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_tag_id(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = outdent(
                """\
                  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
                  * fence-rh-2	(stonith:fence_xvm):	 Stopped
                  * x3	(ocf:pacemaker:Dummy):	 Stopped
                  * y1	(ocf:pacemaker:Dummy):	 Stopped
                """
            )
        else:
            stdout_full = outdent(
                """\
                  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
                  * fence-rh-2	(stonith:fence_xvm):	 Stopped
                  * x3	(ocf::pacemaker:Dummy):	 Stopped
                  * y1	(ocf::pacemaker:Dummy):	 Stopped
                """
            )
        self.assert_pcs_success(
            self.command + ["tag-mixed-stonith-devices-and-resources"],
            stdout_full=stdout_full,
        )

    def test_tag_id_hide_inactive(self):
        self.assert_pcs_success(
            self.command
            + ["tag-mixed-stonith-devices-and-resources", "--hide-inactive"],
            stdout_full=outdent(
                """\
                  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
                """
            ),
        )

    def test_tag_id_with_node(self):
        if is_pacemaker_21_without_20_compatibility():
            stdout_full = outdent(
                """\
                  * fence-rh-2	(stonith:fence_xvm):	 Stopped
                  * x3	(ocf:pacemaker:Dummy):	 Stopped
                  * y1	(ocf:pacemaker:Dummy):	 Stopped
                """
            )
        else:
            stdout_full = outdent(
                """\
                  * fence-rh-2	(stonith:fence_xvm):	 Stopped
                  * x3	(ocf::pacemaker:Dummy):	 Stopped
                  * y1	(ocf::pacemaker:Dummy):	 Stopped
                """
            )
        self.assert_pcs_success(
            self.command
            + ["tag-mixed-stonith-devices-and-resources", "node=rh-2"],
            stdout_full=stdout_full,
        )

    def test_tag_id_with_node_hide_inactive(self):
        self.assert_pcs_success(
            self.command
            + [
                "tag-mixed-stonith-devices-and-resources",
                "node=rh-1",
                "--hide-inactive",
            ],
            stdout_full=outdent(
                """\
                  * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
                """
            ),
        )

    def test_resource_status_without_id(self):
        self.assert_pcs_success(
            self.command, stdout_full=self.all_resources_output
        )

    def test_resource_status_without_id_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["--hide-inactive"],
            stdout_full=self.active_resources_output,
        )

    def test_resource_status_without_id_with_node(self):
        self.assert_pcs_success(
            self.command + ["node=rh-1"], stdout_full=self.node_output
        )

    def test_resource_status_without_id_with_node_hide_inactive(self):
        self.assert_pcs_success(
            self.command + ["node=rh-1", "--hide-inactive"],
            stdout_full=self.active_resources_output_node,
        )

    def test_resource_status_without_id_default_command(self):
        self.assert_pcs_success(
            self.command[:-1], stdout_full=self.all_resources_output
        )

    def test_resource_status_without_id_default_command_hide_inactive(self):
        self.assert_pcs_success(
            self.command[:-1] + ["--hide-inactive"],
            stdout_full=self.active_resources_output,
        )

    def test_status_no_resources(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(self.command, stdout_full=self.no_resources_msg)

    def test_status_no_resources_hide_inactive(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            self.command + ["--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_status_no_resources_with_node(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            self.command + ["node=rh-1"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_status_no_resources_with_node_hide_inactive(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            self.command + ["node=rh-1", "--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )

    def test_status_no_resources_default_command(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            self.command[:-1], stdout_full=self.no_resources_msg
        )

    def test_status_no_resources_default_command_hide_inactive(self):
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            self.command[:-1] + ["--hide-inactive"],
            stdout_full=self.no_active_resources_msg,
        )


class StonithStatus(ResourceStonithStatusBase, TestCase):
    command = ["stonith", "status"]
    no_resources_msg = "NO stonith devices configured\n"
    all_resources_output = outdent(
        """\
          * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
          * fence-rh-2	(stonith:fence_xvm):	 Stopped
          * fence-kdump	(stonith:fence_kdump):	 Stopped

        Fencing Levels:
         Target: rh-1
           Level 1 - fence-kdump
           Level 2 - fence-rh-1
         Target: rh-2
           Level 1 - fence-kdump
           Level 2 - fence-rh-2
        """
    )
    active_resources_output = outdent(
        """\
          * fence-rh-1	(stonith:fence_xvm):	 Started rh-1

        Fencing Levels:
         Target: rh-1
           Level 1 - fence-kdump
           Level 2 - fence-rh-1
         Target: rh-2
           Level 1 - fence-kdump
           Level 2 - fence-rh-2
        """
    )
    active_resources_output_node = outdent(
        """\
          * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
        """
    )
    node_output = outdent(
        """\
          * fence-rh-1	(stonith:fence_xvm):	 Started rh-1
          * fence-rh-2	(stonith:fence_xvm):	 Stopped
          * fence-kdump	(stonith:fence_kdump):	 Stopped
        """
    )


def fixture_resources_status_output(nodes="rh-1 rh-2", inactive=True):
    if not inactive:
        if is_pacemaker_21_without_20_compatibility():
            return outdent(
                """\
                  * x1	(ocf:pacemaker:Dummy):	 Started rh-1
                """
            )
        return outdent(
            """\
              * x1	(ocf::pacemaker:Dummy):	 Started rh-1
            """
        )

    if is_pacemaker_21_without_20_compatibility():
        return outdent(
            f"""\
              * not-in-tags	(ocf:pacemaker:Dummy):	 Stopped
              * x1	(ocf:pacemaker:Dummy):	 Started rh-1
              * x2	(ocf:pacemaker:Dummy):	 Stopped
              * x3	(ocf:pacemaker:Dummy):	 Stopped
              * y1	(ocf:pacemaker:Dummy):	 Stopped
              * Clone Set: y2-clone [y2]:
                * Stopped: [ {nodes} ]
            """
        )
    return outdent(
        f"""\
          * not-in-tags	(ocf::pacemaker:Dummy):	 Stopped
          * x1	(ocf::pacemaker:Dummy):	 Started rh-1
          * x2	(ocf::pacemaker:Dummy):	 Stopped
          * x3	(ocf::pacemaker:Dummy):	 Stopped
          * y1	(ocf::pacemaker:Dummy):	 Stopped
          * Clone Set: y2-clone [y2]:
            * Stopped: [ {nodes} ]
        """
    )


class ResourceStatus(ResourceStonithStatusBase, TestCase):
    command = ["resource", "status"]
    no_resources_msg = "NO resources configured\n"
    all_resources_output = fixture_resources_status_output()
    active_resources_output = fixture_resources_status_output(inactive=False)
    active_resources_output_node = active_resources_output
    node_output = fixture_resources_status_output(nodes="rh-1")


class StatusResources(ResourceStonithStatusBase, TestCase):
    command = ["status", "resources"]
    no_resources_msg = "NO resources configured\n"
    all_resources_output = fixture_resources_status_output()
    active_resources_output = fixture_resources_status_output(inactive=False)
    active_resources_output_node = active_resources_output
    node_output = fixture_resources_status_output(nodes="rh-1")
    no_resources_status = outdent(
        """\
        Cluster name: test99

        WARNINGS:
        No stonith devices and stonith-enabled is not false

        Cluster Summary:
          * Stack: unknown
          * Current DC: NONE
        """
    )
    resources_status = outdent(
        """\
        Cluster name: test99
        Cluster Summary:
          * Stack: unknown
          * Current DC: rh-2 (version unknown) - partition WITHOUT quorum
        """
    )

    def test_resource_status_without_id_default_command(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_success(["status"], stdout_start=self.resources_status)

    def test_status_no_resources_default_command(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            ["status"], stdout_start=self.no_resources_status
        )

    def test_resource_status_without_id_default_command_hide_inactive(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        self.assert_pcs_success(
            ["status", "--hide-inactive"], stdout_start=self.resources_status
        )

    def test_status_no_resources_default_command_hide_inactive(self):
        self.pcs_runner.corosync_conf_opt = self.corosync_conf
        write_file_to_tmpfile(rc("cib-empty.xml"), self.temp_cib)
        self.assert_pcs_success(
            ["status", "--hide-inactive"], stdout_start=self.no_resources_status
        )



from unittest import TestCase

from lxml import etree

from pcs.lib.cib.resource import bundle

# pcs.lib.cib.resource.bundle is covered by:
# - pcs_test.tier0.lib.commands.resource.test_bundle_create
# - pcs_test.tier0.lib.commands.resource.test_bundle_update
# - pcs_test.tier0.lib.commands.resource.test_resource_create


class IsBundle(TestCase):
    def test_is_bundle(self):
        self.assertTrue(bundle.is_bundle(etree.fromstring("<bundle/>")))
        self.assertFalse(bundle.is_bundle(etree.fromstring("<clone/>")))
        self.assertFalse(bundle.is_bundle(etree.fromstring("<group/>")))


class GetInnerResource(TestCase):
    def assert_inner_resource(self, resource_id, xml):
        self.assertEqual(
            resource_id,
            bundle.get_inner_resource(etree.fromstring(xml)).get("id", ""),
        )

    def test_primitive(self):
        self.assert_inner_resource(
            "A",
            """
                <bundle id="B">
                    <meta_attributes />
                    <primitive id="A" />
                    <meta_attributes />
                </bundle>
            """,
        )



from unittest import TestCase

from lxml import etree

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.bin_mock import get_mock_settings
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    outdent,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

# pylint: disable=line-too-long
# pylint: disable=too-many-public-methods

INVALID_CONTAINER_OPTION_MSG = (
    "invalid container option 'extra', allowed options are: "
    "'image', 'masters', 'network', 'options', 'promoted-max', "
    "'replicas', 'replicas-per-host', 'run-command'"
)
ERRORS_HAVE_OCCURRED = (
    "Error: Errors have occurred, therefore pcs is unable to continue\n"
)


class BundleCreateCommon(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    ),
):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_bundle_create")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_binary")

    def tearDown(self):
        self.temp_cib.close()


class BundleReset(BundleCreateCommon):
    empty_cib = rc("cib-empty.xml")

    def test_minimal(self):
        self.assert_pcs_success(
            "resource bundle create B1 container docker image=pcs:test".split()
        )
        self.assert_pcs_success(
            "resource bundle create B2 container docker image=pcs:test".split()
        )
        self.assert_effect(
            "resource bundle reset B1 container image=pcs:new".split(),
            """
                <resources>
                    <bundle id="B1">
                        <docker image="pcs:new" />
                    </bundle>
                    <bundle id="B2">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """,
        )


class BundleCreate(BundleCreateCommon):
    empty_cib = rc("cib-empty.xml")

    def test_minimal(self):
        self.assert_effect(
            "resource bundle create B1 container docker image=pcs:test".split(),
            """
                <resources>
                    <bundle id="B1">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """,
        )

    def test_all_options(self):
        self.assert_effect(
            """
                resource bundle create B1
                container docker replicas=4 replicas-per-host=2
                    run-command=/bin/true
                port-map port=1001
                meta target-role=Stopped
                network control-port=12345 host-interface=eth0 host-netmask=24
                port-map id=B1-port-map-1001 internal-port=2002 port=2000
                port-map range=3000-3300
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                network ip-range-start=192.168.100.200
                storage-map id=B1-storage-map source-dir=/tmp/docker2a
                    target-dir=/tmp/docker2b
                container image=pcs:test promoted-max=0
                meta is-managed=false
                storage-map source-dir-root=/tmp/docker3a
                    target-dir=/tmp/docker3b
                storage-map id=B1-port-map-1001-1 source-dir-root=/tmp/docker4a
                    target-dir=/tmp/docker4b
                container network=extra_network_settings options=extra_options
            """.split(),
            """
                <resources>
                    <bundle id="B1">
                        <docker
                            image="pcs:test"
                            network="extra_network_settings"
                            options="extra_options"
                            promoted-max="0"
                            replicas="4"
                            replicas-per-host="2"
                            run-command="/bin/true"
                        />
                        <network
                            control-port="12345"
                            host-interface="eth0"
                            host-netmask="24"
                            ip-range-start="192.168.100.200"
                        >
                            <port-mapping id="B1-port-map-1001-2" port="1001" />
                            <port-mapping
                                id="B1-port-map-1001"
                                internal-port="2002"
                                port="2000"
                            />
                            <port-mapping
                                id="B1-port-map-3000-3300"
                                range="3000-3300"
                            />
                        </network>
                        <storage>
                            <storage-mapping
                                id="B1-storage-map-1"
                                source-dir="/tmp/docker1a"
                                target-dir="/tmp/docker1b"
                            />
                            <storage-mapping
                                id="B1-storage-map"
                                source-dir="/tmp/docker2a"
                                target-dir="/tmp/docker2b"
                            />
                            <storage-mapping
                                id="B1-storage-map-2"
                                source-dir-root="/tmp/docker3a"
                                target-dir="/tmp/docker3b"
                            />
                            <storage-mapping
                                id="B1-port-map-1001-1"
                                source-dir-root="/tmp/docker4a"
                                target-dir="/tmp/docker4b"
                            />
                        </storage>
                        <meta_attributes id="B1-meta_attributes">
                            <nvpair
                                id="B1-meta_attributes-is-managed"
                                name="is-managed"
                                value="false"
                            />
                            <nvpair
                                id="B1-meta_attributes-target-role"
                                name="target-role"
                                value="Stopped"
                            />
                        </meta_attributes>
                    </bundle>
                </resources>
            """,
        )

    def test_deprecated_masters(self):
        self.assert_effect(
            """
                resource bundle create B1
                container docker image=pcs:test masters=0
            """.split(),
            """
                <resources>
                    <bundle id="B1">
                        <docker image="pcs:test" masters="0" />
                    </bundle>
                </resources>
            """,
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )

    def test_deprecated_masters_and_promoted_max(self):
        self.assert_pcs_fail(
            """
                resource bundle create B1
                container docker image=pcs:test masters=0 promoted-max=0
            """.split(),
            "Error: Only one of container options 'masters' and 'promoted-max' "
            "can be used\n"
            "Warning: container option 'masters' is deprecated and should "
            "not be used, use 'promoted-max' instead\n" + ERRORS_HAVE_OCCURRED,
        )

    def test_fail_when_missing_args_1(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle".split(),
            stdout_start="\nUsage: pcs resource bundle...\n",
        )

    def test_fail_when_missing_args_2(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create".split(),
            stdout_start="\nUsage: pcs resource bundle create...\n",
        )

    def test_fail_when_missing_container_type(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B1".split(),
            "Error: '' is not a valid container type value, use 'docker', "
            "'podman', 'rkt'\n" + ERRORS_HAVE_OCCURRED,
        )

    def test_fail_when_missing_required(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B1 container docker".split(),
            "Error: required container option 'image' is missing\n"
            + ERRORS_HAVE_OCCURRED,
        )

    def test_fail_on_unknown_option(self):
        self.assert_pcs_fail(
            """
                resource bundle create B1 container docker image=pcs:test
                extra=option
            """.split(),
            f"Error: {INVALID_CONTAINER_OPTION_MSG}, use --force to override\n"
            + ERRORS_HAVE_OCCURRED,
        )

    def test_unknown_option_forced(self):
        # Test that pcs allows to specify options it does not know about. This
        # ensures some kind of forward compatibility, so the user will be able
        # to specify new options. However, as of now the option is not
        # supported by pacemaker and so the command fails.
        self.assert_pcs_fail(
            """
                resource bundle create B1 container docker image=pcs:test
                extra=option --force
            """.split(),
            stdout_start=f"Warning: {INVALID_CONTAINER_OPTION_MSG}\nError: Unable to update cib\n",
        )

    def test_more_errors(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B#1 container docker replicas=x".split(),
            outdent(
                """\
                Error: invalid bundle name 'B#1', '#' is not a valid character for a bundle name
                Error: required container option 'image' is missing
                Error: 'x' is not a valid replicas value, use a positive integer
                """
            )
            + ERRORS_HAVE_OCCURRED,
        )

    def assert_no_options(self, keyword):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B".split() + [keyword],
            "Error: No {0} options specified\n".format(keyword),
        )

    def test_empty_container(self):
        self.assert_no_options("container")

    def test_empty_network(self):
        self.assert_no_options("network")

    def test_empty_storage_map(self):
        self.assert_no_options("storage-map")

    def test_empty_port_map(self):
        self.assert_no_options("port-map")

    def test_empty_meta(self):
        self.assert_no_options("meta")


class BundleUpdate(BundleCreateCommon):
    empty_cib = rc("cib-empty.xml")

    success_xml = """
        <resources>
            <bundle id="B">
                <docker
                    image="pcs:test"
                    replicas="6"
                    replicas-per-host="2"
                />
                <network
                    host-interface="eth1"
                    host-netmask="24"
                    ip-range-start="192.168.100.200"
                >
                    <port-mapping
                        id="B-port-map-2001"
                        internal-port="1001"
                        port="2001"
                    />
                    <port-mapping
                        id="B-port-map-2003"
                        internal-port="1003"
                        port="2003"
                    />
                </network>
                <storage>
                    <storage-mapping
                        id="B-storage-map-1"
                        source-dir="/tmp/docker2a"
                        target-dir="/tmp/docker2b"
                    />
                    <storage-mapping
                        id="B-storage-map"
                        source-dir="/tmp/docker4a"
                        target-dir="/tmp/docker4b"
                    />
                </storage>
                <meta_attributes id="B-meta_attributes">
                    <nvpair id="B-meta_attributes-priority"
                        name="priority" value="10" />
                    <nvpair id="B-meta_attributes-resource-stickiness"
                        name="resource-stickiness" value="100" />
                    <nvpair id="B-meta_attributes-target-role"
                        name="target-role" value="Stopped" />
                </meta_attributes>
            </bundle>
        </resources>
    """

    @staticmethod
    def success_command(remove_command):
        return [
            "resource",
            "bundle",
            "update",
            "B",
            "container",
            "promoted-max=",
            "replicas=6",
            "replicas-per-host=2",
            "network",
            "control-port=",
            "host-interface=eth1",
            "ip-range-start=192.168.100.200",
            "port-map",
            remove_command,
            "B-port-map-2000",
            "B-port-map-2002",
            "port-map",
            "add",
            "internal-port=1003",
            "port=2003",
            "storage-map",
            remove_command,
            "B-storage-map",
            "B-storage-map-2",
            "storage-map",
            "add",
            "source-dir=/tmp/docker4a",
            "target-dir=/tmp/docker4b",
            "meta",
            "priority=10",
            "is-managed=",
            "target-role=Stopped",
        ]

    def fixture_bundle(self, name):
        self.assert_pcs_success(
            [
                "resource",
                "bundle",
                "create",
                name,
                "container",
                "docker",
                "image=pcs:test",
            ]
        )

    def fixture_bundle_complex(self, name):
        self.assert_pcs_success(
            [
                "resource",
                "bundle",
                "create",
                name,
                "container",
                "docker",
                "image=pcs:test",
                "replicas=4",
                "promoted-max=2",
                "network",
                "control-port=12345",
                "host-interface=eth0",
                "host-netmask=24",
                "port-map",
                "internal-port=1000",
                "port=2000",
                "port-map",
                "internal-port=1001",
                "port=2001",
                "port-map",
                "internal-port=1002",
                "port=2002",
                "storage-map",
                "source-dir=/tmp/docker1a",
                "target-dir=/tmp/docker1b",
                "storage-map",
                "source-dir=/tmp/docker2a",
                "target-dir=/tmp/docker2b",
                "storage-map",
                "source-dir=/tmp/docker3a",
                "target-dir=/tmp/docker3b",
                "meta",
                "priority=15",
                "resource-stickiness=100",
                "is-managed=false",
            ]
        )

    def test_fail_when_missing_args_1(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update".split(),
            stdout_start="\nUsage: pcs resource bundle update...\n",
        )

    def test_fail_when_missing_args_2(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B port-map".split(),
            "Error: No port-map options specified\n",
        )

    def test_fail_when_missing_args_3(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B storage-map remove".split(),
            "Error: When using 'storage-map' you must specify either 'add' and "
            "options or either of 'delete' or 'remove' and id(s)\n",
        )

    def test_fail_when_missing_args_4(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B storage-map delete".split(),
            "Error: When using 'storage-map' you must specify either 'add' and "
            "options or either of 'delete' or 'remove' and id(s)\n",
        )

    def test_bad_id(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B1 container image=test".split(),
            "Error: bundle 'B1' does not exist\n",
        )

    def test_success_delete(self):
        self.fixture_bundle_complex("B")
        self.assert_effect(
            self.success_command(remove_command="delete"),
            self.success_xml,
        )

    def test_success_remove(self):
        self.fixture_bundle_complex("B")
        self.assert_effect(
            self.success_command(remove_command="remove"),
            self.success_xml,
        )

    def test_deprecated_masters_set(self):
        # Setting both deprecated options and their new variants is tested in
        # self.test_options_errors. This shows deprecated options emit warning
        # even when not forced.
        self.fixture_bundle("B")
        self.assert_effect(
            "resource bundle update B container masters=2".split(),
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" masters="2" />
                    </bundle>
                </resources>
            """,
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )

    def test_delete_masters(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container masters=2".split(),
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )
        self.assert_effect(
            "resource bundle update B container masters=".split(),
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """,
        )

    def test_delete_masters_and_promoted_max(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container masters= promoted-max=".split(),
        )

    def test_masters_set_after_promoted_max(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container promoted-max=3".split(),
        )
        self.assert_pcs_fail(
            "resource bundle update B container masters=2".split(),
            "Warning: container option 'masters' is deprecated and should "
            "not be used, use 'promoted-max' instead\n"
            "Error: Cannot set container option 'masters' because container "
            "option 'promoted-max' is already set\n" + ERRORS_HAVE_OCCURRED,
        )

    def test_masters_set_after_promoted_max_with_remove(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container promoted-max=3".split(),
        )
        self.assert_effect(
            "resource bundle update B container masters=2 promoted-max=".split(),
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" masters="2" />
                    </bundle>
                </resources>
            """,
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )

    def test_promoted_max_set_after_masters(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container masters=2".split(),
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )
        self.assert_pcs_fail(
            "resource bundle update B container promoted-max=3".split(),
            "Error: Cannot set container option 'promoted-max' because "
            "container option 'masters' is already set\n"
            + ERRORS_HAVE_OCCURRED,
        )

    def test_promoted_max_set_after_masters_with_remove(self):
        self.fixture_bundle("B")
        self.assert_pcs_success(
            "resource bundle update B container masters=2".split(),
            "Warning: container option 'masters' is deprecated and should not "
            "be used, use 'promoted-max' instead\n",
        )
        self.assert_effect(
            "resource bundle update B container masters= promoted-max=3".split(),
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" promoted-max="3" />
                    </bundle>
                </resources>
            """,
        )

    def test_force_unknown_option(self):
        self.fixture_bundle("B")

        self.assert_pcs_fail(
            "resource bundle update B container extra=option".split(),
            f"Error: {INVALID_CONTAINER_OPTION_MSG}, use --force "
            "to override\n" + ERRORS_HAVE_OCCURRED,
        )
        # Test that pcs allows to specify options it does not know about. This
        # ensures some kind of forward compatibility, so the user will be able
        # to specify new options. However, as of now the option is not
        # supported by pacemaker and so the command fails.
        self.assert_pcs_fail(
            "resource bundle update B container extra=option --force".split(),
            stdout_start=f"Warning: {INVALID_CONTAINER_OPTION_MSG}\nError: Unable to update cib\n",
        )

        # no force needed when removing an unknown option
        self.assert_effect(
            "resource bundle update B container extra=".split(),
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """,
        )

    def assert_no_options(self, keyword):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B".split() + [keyword],
            "Error: No {0} options specified\n".format(keyword),
        )

    def test_empty_container(self):
        self.assert_no_options("container")

    def test_empty_network(self):
        self.assert_no_options("network")

    def test_empty_storage_map(self):
        self.assert_no_options("storage-map")

    def test_empty_port_map(self):
        self.assert_no_options("port-map")

    def test_empty_meta(self):
        self.assert_no_options("meta")


class BundleShow(TestCase, AssertPcsMixin):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_bundle_show")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        self.pcs_runner.mock_settings = get_mock_settings("crm_resource_binary")

    def tearDown(self):
        self.temp_cib.close()

    def test_minimal(self):
        self.assert_pcs_success(
            "resource bundle create B1 container docker image=pcs:test".split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
            """
            ),
        )

    def test_container(self):
        self.assert_pcs_success(
            [
                "resource",
                "bundle",
                "create",
                "B1",
                "container",
                "docker",
                "image=pcs:test",
                "promoted-max=2",
                "replicas=4",
                "options=a b c",
            ]
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test replicas=4 promoted-max=2 options="a b c"
            """
            ),
        )

    def test_network(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                network host-interface=eth0 host-netmask=24 control-port=12345
            """.split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
              Network: control-port=12345 host-interface=eth0 host-netmask=24
            """
            ),
        )

    def test_port_map(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                port-map id=B1-port-map-1001 internal-port=2002 port=2000
                port-map range=3000-3300
            """.split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
              Port Mapping:
                port=2000 internal-port=2002 (B1-port-map-1001)
                range=3000-3300 (B1-port-map-3000-3300)
            """
            ),
        )

    def test_storage_map(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                storage-map id=my-storage-map source-dir=/tmp/docker2a
                    target-dir=/tmp/docker2b
            """.split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
              Storage Mapping:
                source-dir=/tmp/docker1a target-dir=/tmp/docker1b (B1-storage-map)
                source-dir=/tmp/docker2a target-dir=/tmp/docker2b (my-storage-map)
            """
            ),
        )

    def test_meta(self):
        self.assert_pcs_success(
            """
            resource bundle create B1 container docker image=pcs:test
            --disabled
            """.split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
              Meta Attributes: B1-meta_attributes
                target-role=Stopped
            """
            ),
        )

    def test_resource(self):
        self.assert_pcs_success(
            (
                "resource bundle create B1 "
                "container docker image=pcs:test "
                "network control-port=1234"
            ).split()
        )
        self.assert_pcs_success(
            "resource create A ocf:pacemaker:Dummy bundle B1 --no-default-ops".split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test
              Network: control-port=1234
              Resource: A (class=ocf provider=pacemaker type=Dummy)
                Operations:
                  monitor: A-monitor-interval-10s
                    interval=10s
                    timeout=20s
            """
            ),
        )

    def test_all(self):
        self.assert_pcs_success(
            [
                "resource",
                "bundle",
                "create",
                "B1",
                "container",
                "docker",
                "image=pcs:test",
                "promoted-max=2",
                "replicas=4",
                "options=a b c",
                "network",
                "host-interface=eth0",
                "host-netmask=24",
                "control-port=12345",
                "port-map",
                "id=B1-port-map-1001",
                "internal-port=2002",
                "port=2000",
                "port-map",
                "range=3000-3300",
                "storage-map",
                "source-dir=/tmp/docker1a",
                "target-dir=/tmp/docker1b",
                "storage-map",
                "id=my-storage-map",
                "source-dir=/tmp/docker2a",
                "target-dir=/tmp/docker2b",
                "meta",
                "target-role=Stopped",
                "is-managed=false",
            ]
        )
        self.assert_pcs_success(
            "resource create A ocf:pacemaker:Dummy bundle B1 --no-default-ops".split()
        )
        self.assert_pcs_success(
            "resource config B1".split(),
            outdent(
                """\
            Bundle: B1
              Docker: image=pcs:test replicas=4 promoted-max=2 options="a b c"
              Network: control-port=12345 host-interface=eth0 host-netmask=24
              Port Mapping:
                port=2000 internal-port=2002 (B1-port-map-1001)
                range=3000-3300 (B1-port-map-3000-3300)
              Storage Mapping:
                source-dir=/tmp/docker1a target-dir=/tmp/docker1b (B1-storage-map)
                source-dir=/tmp/docker2a target-dir=/tmp/docker2b (my-storage-map)
              Meta Attributes: B1-meta_attributes
                is-managed=false
                target-role=Stopped
              Resource: A (class=ocf provider=pacemaker type=Dummy)
                Operations:
                  monitor: A-monitor-interval-10s
                    interval=10s
                    timeout=20s
            """
            ),
        )



from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from pcs.lib.commands.resource_agent import (
    _agent_metadata_to_dict,
    _complete_agent_list,
    _get_agent_names,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.resource_agent import (
    InvalidResourceAgentName,
    ResourceAgentError,
    ResourceAgentFacadeFactory,
    ResourceAgentName,
    StandardProviderTuple,
    resource_agent_error_to_report_item,
)


# deprecated: use pcs.commands.resource_agent.get_agent_list instead
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all stonith agents on the local host, optionally filtered and described

    describe -- load and return agents' description as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()
    return _complete_agent_list(
        runner,
        lib_env.report_processor,
        sorted(
            _get_agent_names(runner, StandardProviderTuple("stonith")),
            key=lambda item: item.full_name,
        ),
        describe,
        search,
    )


# deprecated: use pcs.commands.resource_agent.get_agent_metadata instead
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent (not containing "stonith:" prefix)
    """
    runner = lib_env.cmd_runner()
    agent_factory = ResourceAgentFacadeFactory(runner, lib_env.report_processor)
    try:
        if ":" in agent_name:
            raise InvalidResourceAgentName(agent_name)
        return _agent_metadata_to_dict(
            agent_factory.facade_from_parsed_name(
                ResourceAgentName("stonith", None, agent_name)
            ).metadata,
            describe=True,
        )
    except ResourceAgentError as e:
        lib_env.report_processor.report(
            resource_agent_error_to_report_item(e, is_stonith=True)
        )
        raise LibraryError() from e



from unittest import (
    TestCase,
    mock,
)

from pcs import stonith
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import InputModifiers
from pcs.common import reports

from pcs_test.tools.misc import dict_to_modifiers


def _dict_to_modifiers(options):
    def _convert_val(val):
        if val is True:
            return ""
        return val

    return InputModifiers(
        {
            f"--{opt}": _convert_val(val)
            for opt, val in options.items()
            if val is not False
        }
    )


class SbdEnable(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["sbd"])
        self.sbd = mock.Mock(spec_set=["enable_sbd"])
        self.lib.sbd = self.sbd

    def assert_called_with(
        self, default_watchdog, watchdog_dict, sbd_options, **kwargs
    ):
        default_kwargs = dict(
            default_device_list=None,
            node_device_dict=None,
            allow_unknown_opts=False,
            ignore_offline_nodes=False,
            no_watchdog_validation=False,
        )
        default_kwargs.update(kwargs)
        self.sbd.enable_sbd.assert_called_once_with(
            default_watchdog, watchdog_dict, sbd_options, **default_kwargs
        )

    def call_cmd(self, argv, modifiers=None):
        stonith.sbd_enable(self.lib, argv, _dict_to_modifiers(modifiers or {}))

    def test_no_args(self):
        self.call_cmd([])
        self.assert_called_with(None, {}, {})

    def test_watchdog(self):
        self.call_cmd(["watchdog=/dev/wd"])
        self.assert_called_with("/dev/wd", {}, {})

    def test_device(self):
        self.call_cmd(["device=/dev/sda"])
        self.assert_called_with(None, {}, {}, default_device_list=["/dev/sda"])

    def test_options(self):
        self.call_cmd(["SBD_A=a", "SBD_B=b"])
        self.assert_called_with(None, {}, dict(SBD_A="a", SBD_B="b"))

    def test_multiple_watchdogs_devices(self):
        self.call_cmd(
            [
                "watchdog=/dev/wd",
                "watchdog=/dev/wda@node-a",
                "watchdog=/dev/wdb@node-b",
                "device=/dev/sda1",
                "device=/dev/sda2",
                "device=/dev/sdb1@node-b",
                "device=/dev/sdb2@node-b",
                "device=/dev/sdc1@node-c",
                "device=/dev/sdc2@node-c",
            ]
        )
        self.assert_called_with(
            "/dev/wd",
            {"node-a": "/dev/wda", "node-b": "/dev/wdb"},
            {},
            default_device_list=["/dev/sda1", "/dev/sda2"],
            node_device_dict={
                "node-b": ["/dev/sdb1", "/dev/sdb2"],
                "node-c": ["/dev/sdc1", "/dev/sdc2"],
            },
        )

    def test_modifiers(self):
        self.call_cmd(
            [],
            modifiers={
                "force": "",
                "skip-offline": "",
                "no-watchdog-validation": "",
            },
        )
        self.assert_called_with(
            None,
            {},
            {},
            allow_unknown_opts=True,
            ignore_offline_nodes=True,
            no_watchdog_validation=True,
        )


class SbdDeviceSetup(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["sbd"])
        self.sbd = mock.Mock(spec_set=["initialize_block_devices"])
        self.lib.sbd = self.sbd

    def assert_called_with(self, device_list, option_dict):
        self.sbd.initialize_block_devices.assert_called_once_with(
            device_list, option_dict
        )

    def call_cmd(self, argv, modifiers=None):
        all_modifiers = dict(
            force=True,  # otherwise it asks interactively for confirmation
        )
        all_modifiers.update(modifiers or {})
        stonith.sbd_setup_block_device(
            self.lib, argv, _dict_to_modifiers(all_modifiers)
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertEqual(cm.exception.message, "No device defined")

    def test_minimal(self):
        self.call_cmd(["device=/dev/sda"])
        self.assert_called_with(["/dev/sda"], {})

    def test_devices_and_options(self):
        self.call_cmd(["device=/dev/sda", "a=A", "device=/dev/sdb", "b=B"])
        self.assert_called_with(["/dev/sda", "/dev/sdb"], {"a": "A", "b": "B"})

    def test_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["a=A"])
        self.assertEqual(cm.exception.message, "No device defined")


class StonithUpdateScsiDevices(TestCase):
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.lib = mock.Mock(spec_set=["stonith"])
        self.stonith = mock.Mock(
            spec_set=["update_scsi_devices", "update_scsi_devices_add_remove"]
        )
        self.lib.stonith = self.stonith

    def assert_called_with(self, stonith_id, set_devices, force_flags):
        self.stonith.update_scsi_devices.assert_called_once_with(
            stonith_id, set_devices, force_flags=force_flags
        )
        self.stonith.update_scsi_devices_add_remove.assert_not_called()

    def assert_add_remove_called_with(
        self, stonith_id, add_devices, remove_devices, force_flags
    ):
        self.stonith.update_scsi_devices_add_remove.assert_called_once_with(
            stonith_id, add_devices, remove_devices, force_flags=force_flags
        )
        self.stonith.update_scsi_devices.assert_not_called()

    def assert_bad_syntax_cli_exception(self, args):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(args)
        self.assertEqual(cm.exception.message, None)
        self.assertEqual(
            cm.exception.hint,
            (
                "You must specify either list of set devices or at least one "
                "device for add or delete/remove devices"
            ),
        )
        self.stonith.update_scsi_devices.assert_not_called()
        self.stonith.update_scsi_devices_add_remove.assert_not_called()

    def call_cmd(self, argv, modifiers=None):
        stonith.stonith_update_scsi_devices(
            self.lib, argv, dict_to_modifiers(modifiers or {})
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd([])
        self.assertEqual(cm.exception.message, None)

    def test_only_stonith_id(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["stonith-id"])
        self.assertEqual(cm.exception.message, None)

    def test_unknown_keyword(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["stonith-id", "unset"])
        self.assertEqual(cm.exception.message, None)

    def test_supported_options(self):
        self.call_cmd(
            ["stonith-id", "set", "d1", "d2"],
            {"skip-offline": True, "request-timeout": 60},
        )
        self.assert_called_with(
            "stonith-id",
            ["d1", "d2"],
            [reports.codes.SKIP_OFFLINE_NODES],
        )

    def test_unsupported_options(self):
        with self.assertRaises(CmdLineInputError) as cm:
            self.call_cmd(["stonith-id", "set", "d1", "d2"], {"force": True})
        self.assertEqual(
            cm.exception.message,
            "Specified option '--force' is not supported in this command",
        )

    def test_only_set_keyword(self):
        self.assert_bad_syntax_cli_exception(["stonith-id", "set"])

    def test_only_add_keyword(self):
        self.assert_bad_syntax_cli_exception(["stonith-id", "add"])

    def test_only_remove_keyword(self):
        self.assert_bad_syntax_cli_exception(["stonith-id", "remove"])

    def test_only_delete_keyword(self):
        self.assert_bad_syntax_cli_exception(["stonith-id", "delete"])

    def test_add_and_empty_remove(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "add", "d1", "remove"]
        )

    def test_add_and_empty_delete(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "add", "d1", "delete"]
        )

    def test_empty_add_and_remove(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "add", "remove", "d1"]
        )

    def test_empty_add_and_delete(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "add", "delete", "d1"]
        )

    def test_empty_remove_and_delete(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "remove", "delete", "d1"]
        )

    def test_empty_delete_and_remove(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "delete", "remove", "d1"]
        )

    def test_empty_add_empty_remove_empty_delete(self):
        self.assert_bad_syntax_cli_exception(
            ["stonith-id", "add", "delete", "remove"]
        )

    def test_set_add_remove_delete_devices(self):
        self.assert_bad_syntax_cli_exception(
            [
                "stonith-id",
                "set",
                "add",
                "d2",
                "remove",
                "d3",
                "delete",
                "d4",
            ]
        )

    def test_set_devices(self):
        self.call_cmd(["stonith-id", "set", "d1", "d2"])
        self.assert_called_with("stonith-id", ["d1", "d2"], [])

    def test_add_devices(self):
        self.call_cmd(["stonith-id", "add", "d1", "d2"])
        self.assert_add_remove_called_with("stonith-id", ["d1", "d2"], [], [])

    def test_remove_devices(self):
        self.call_cmd(["stonith-id", "remove", "d1", "d2"])
        self.assert_add_remove_called_with("stonith-id", [], ["d1", "d2"], [])

    def test_delete_devices(self):
        self.call_cmd(["stonith-id", "delete", "d1", "d2"])
        self.assert_add_remove_called_with("stonith-id", [], ["d1", "d2"], [])

    def test_add_remove_devices(self):
        self.call_cmd(["stonith-id", "add", "d1", "d2", "remove", "d3", "d4"])
        self.assert_add_remove_called_with(
            "stonith-id", ["d1", "d2"], ["d3", "d4"], []
        )

    def test_add_delete_devices(self):
        self.call_cmd(["stonith-id", "add", "d1", "d2", "delete", "d3", "d4"])
        self.assert_add_remove_called_with(
            "stonith-id", ["d1", "d2"], ["d3", "d4"], []
        )

    def test_add_delete_remove_devices(self):
        self.call_cmd(
            [
                "stonith-id",
                "add",
                "d1",
                "d2",
                "delete",
                "d3",
                "d4",
                "remove",
                "d5",
            ]
        )
        self.assert_add_remove_called_with(
            "stonith-id", ["d1", "d2"], ["d3", "d4", "d5"], []
        )

    def test_remove_delete_devices(self):
        self.call_cmd(
            ["stonith-id", "remove", "d2", "d1", "delete", "d4", "d3"]
        )
        self.assert_add_remove_called_with(
            "stonith-id", [], ["d4", "d3", "d2", "d1"], []
        )



from unittest import TestCase

import pcs.lib.corosync.config_facade as lib
from pcs.lib.corosync import constants
from pcs.lib.corosync.config_parser import Parser


class GetSimpleValueMixin:
    def assert_value(self, value, config):
        facade = lib.ConfigFacade(Parser.parse(config.encode("utf-8")))
        self.assertEqual(value, self.getter(facade))
        self.assertFalse(facade.need_stopped_cluster)
        self.assertFalse(facade.need_qdevice_reload)


class GetClusterNameTest(GetSimpleValueMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_cluster_name()

    def test_no_name(self):
        self.assert_value("", "")

    def test_empty_name(self):
        self.assert_value("", "totem {\n cluster_name:\n}\n")

    def test_one_name(self):
        self.assert_value("test", "totem {\n cluster_name: test\n}\n")

    def test_more_names(self):
        self.assert_value(
            "TEST", "totem {\n cluster_name: test\n cluster_name: TEST\n}\n"
        )

    def test_more_sections(self):
        self.assert_value(
            "TEST",
            "totem{\n cluster_name: test\n}\ntotem{\n cluster_name: TEST\n}\n",
        )


class GetTransport(GetSimpleValueMixin, TestCase):
    @staticmethod
    def getter(facade):
        return facade.get_transport()

    def test_no_name(self):
        self.assert_value(constants.TRANSPORT_DEFAULT, "")

    def test_empty_name(self):
        self.assert_value(
            constants.TRANSPORT_DEFAULT, "totem {\n transport:\n}\n"
        )

    def test_one_name(self):
        self.assert_value("udp", "totem {\n transport: udp\n}\n")

    def test_more_names(self):
        self.assert_value(
            "udpu", "totem {\n transport: udp\n transport: udpu\n}\n"
        )

    def test_more_sections(self):
        self.assert_value(
            "udpu", "totem{\n transport: udp\n}\ntotem{\n transport: udpu\n}\n"
        )


class GetIpVersion(GetSimpleValueMixin, TestCase):
    transport_ip_list = [
        ("udp", constants.IP_VERSION_4),
        ("udpu", constants.IP_VERSION_64),
        ("knet", constants.IP_VERSION_64),
    ]

    @staticmethod
    def getter(facade):
        return facade.get_ip_version()

    def test_empty_config(self):
        self.assert_value(constants.IP_VERSION_64, "")

    def test_no_name(self):
        for transport, ip in self.transport_ip_list:
            with self.subTest(transport=transport, ip=ip):
                self.assert_value(
                    ip, f"totem {{\n transport: {transport}\n}}\n"
                )

    def test_no_value(self):
        for transport, ip in self.transport_ip_list:
            with self.subTest(transport=transport, ip=ip):
                self.assert_value(
                    ip, f"totem {{\n transport: {transport}\n ip_version:\n}}\n"
                )

    def test_one_name(self):
        self.assert_value(
            "ipv4-6", "totem {\n transport: udp\n ip_version: ipv4-6\n}\n"
        )

    def test_more_names(self):
        self.assert_value(
            "ipv6",
            (
                "totem {\ntransport: udp\nip_version: ipv4-6\nip_version: "
                "ipv6\n}\n"
            ),
        )

    def test_more_sections(self):
        self.assert_value(
            "ipv6",
            "totem {\n transport: knet\n ip_version: ipv4-6\n}\n"
            "totem {\n ip_version: ipv6\n}\n",
        )



from pcs import (
    prop,
    usage,
)
from pcs.cli.common.routing import create_router

property_cmd = create_router(
    {
        "help": lambda _lib, _argv, _modifiers: usage.property(_argv),
        "set": prop.set_property,
        "unset": prop.unset_property,
        # TODO remove, deprecated command
        # replaced with 'config'
        "list": prop.list_property_deprecated,
        # TODO remove, deprecated command
        # replaced with 'config'
        "show": prop.list_property_deprecated,
        "config": prop.list_property,
        "get_cluster_properties_definition": (
            prop.print_cluster_properties_definition
        ),
    },
    ["property"],
    default_cmd="config",
)



from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

CIB_EPOCH_TEMPLATE = """
    <cib epoch="{epoch}" num_updates="122" admin_epoch="0"
        validate-with="pacemaker-3.2" crm_feature_set="3.1.0"
        update-origin="rh7-3" update-client="crmd" cib-last-written="Thu Aug 23
        16:49:17 2012" have-quorum="0" dc-uuid="2"
    >
      <configuration>
        <crm_config/>
        <nodes>
        </nodes>
        <resources/>
        <constraints/>
      </configuration>
      <status/>
    </cib>
"""

CIB_EPOCH = CIB_EPOCH_TEMPLATE.format(epoch=500)
CIB_EPOCH_OLDER = CIB_EPOCH_TEMPLATE.format(epoch=499)
CIB_EPOCH_NEWER = CIB_EPOCH_TEMPLATE.format(epoch=501)

ERROR_CIB_OLD = (
    "Error: Unable to push to the CIB because pushed configuration is older "
    "than existing one. If you are sure you want to push this configuration, "
    "try to use --config to replace only configuration part instead of whole "
    "CIB. Otherwise get current configuration by running command 'pcs cluster "
    "cib' and update that.\n"
)


class CibPush(AssertPcsMixin, TestCase):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_cluster_cib_push")
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        write_data_to_tmpfile(CIB_EPOCH, self.temp_cib)

        self.updated_cib = get_tmp_file("tier1_cluster_cib_push_updated")
        write_data_to_tmpfile("<cib/>", self.updated_cib)
        self.cib_push_cmd = f"cluster cib-push {self.updated_cib.name}".split()
        self.cib_push_diff_cmd = self.cib_push_cmd + [
            f"diff-against={self.temp_cib.name}"
        ]

    def tearDown(self):
        self.temp_cib.close()
        self.updated_cib.close()

    def assert_unable_to_diff(self, reason):
        self.assert_pcs_fail(
            self.cib_push_diff_cmd,
            "Error: unable to diff against original cib '{0}': {1}\n".format(
                self.temp_cib.name, reason
            ),
        )

    def test_cib_too_old(self):
        write_data_to_tmpfile(CIB_EPOCH_OLDER, self.updated_cib)
        self.assert_pcs_fail(self.cib_push_cmd, ERROR_CIB_OLD)

    def test_bad_args(self):
        self.assert_pcs_fail(
            "cluster cib-push a b c".split(), stdout_start="\nUsage: "
        )

    def test_error_scope_and_config(self):
        self.assert_pcs_fail(
            "cluster cib-push file scope=configuration --config".split(),
            "Error: Cannot use both scope and --config\n",
        )

    def test_error_invalid_scope(self):
        self.assert_pcs_fail(
            "cluster cib-push file scope=invalid-scope".split(),
            "Error: invalid CIB scope 'invalid-scope'\n",
        )

    def test_error_scope_and_diff(self):
        self.assert_pcs_fail(
            "cluster cib-push file scope=configuration diff-against=f".split(),
            stdout_start="\nUsage: ",
        )

    def test_error_diff_and_config(self):
        self.assert_pcs_fail(
            "cluster cib-push file diff-against=f --config".split(),
            "Error: Cannot use both scope and diff-against\n",
        )

    def test_error_unknown_option(self):
        self.assert_pcs_fail(
            "cluster cib-push file unknown=value".split(),
            stdout_start="\nUsage: ",
        )

    def test_error_scope_not_present(self):
        self.assert_pcs_fail(
            self.cib_push_cmd + ["--config"],
            (
                "Error: unable to push cib, scope 'configuration' not present "
                "in new cib\n"
            ),
        )

    def test_unable_to_parse_new_cib(self):
        write_data_to_tmpfile("", self.updated_cib)
        self.assert_pcs_fail(
            self.cib_push_cmd, stdout_start="Error: unable to parse new cib:"
        )

    def test_diff_no_difference(self):
        write_data_to_tmpfile(CIB_EPOCH, self.updated_cib)
        self.assert_pcs_success(
            self.cib_push_diff_cmd,
            "The new CIB is the same as the original CIB, nothing to push.\n",
        )

    def test_cib_updated(self):
        write_data_to_tmpfile(CIB_EPOCH_NEWER, self.updated_cib)
        self.assert_pcs_success(self.cib_push_cmd, "CIB updated\n")
